#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import inspect
import os
import tempfile

from typing import Optional

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLLECTION_RESULT_FILE_NAME,
    DOT_IPYNB_EXTENSION,
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
    UNKNOWN_LINE_OF_CODE,
    UNKNOWN_SOURCE_FILE,
)


def get_output_file_path(out_path: Optional[str] = None) -> str:
    """Get the output file path.

    Args:
        out_path (Optional[str], optional): the output path. Defaults to None.

    Returns:
        str: returns the output file path.

    """
    output_directory_path = get_output_directory_path(out_path)
    output_file_path = os.path.join(output_directory_path, COLLECTION_RESULT_FILE_NAME)
    return output_file_path


def get_relative_file_path(path: str) -> str:
    """Get the relative file path.

    Args:
        path (str): a file path.

    Returns:
        str: returns the relative file path of the given file.

    """
    relative_file_path = os.path.relpath(path)
    return relative_file_path


def get_output_directory_path(output_path: Optional[str] = None) -> str:
    """Get the output directory path.

    Returns:
        str: returns the output directory path.

    """
    current_working_directory_path = output_path if output_path else os.getcwd()
    checkpoints_output_directory_path = os.path.join(
        current_working_directory_path, SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME
    )
    os.makedirs(checkpoints_output_directory_path, exist_ok=True)
    return checkpoints_output_directory_path


def get_collection_point_source_file_path() -> str:
    """Get the path of the source file where collection point it is.

    Returns:
        str: returns the path of source file where collection point it is.

    """
    try:
        collection_point_file_path = inspect.stack()[2].filename
        is_temporal_file_path = _is_temporal_path(collection_point_file_path)
        if is_temporal_file_path:
            ipynb_file_path_collection = _get_ipynb_file_path_collection()
            if len(ipynb_file_path_collection) == 1:
                collection_point_file_path = ipynb_file_path_collection[0]
            else:
                collection_point_file_path = UNKNOWN_SOURCE_FILE

        return collection_point_file_path

    except Exception:
        return UNKNOWN_SOURCE_FILE


def get_collection_point_line_of_code() -> int:
    """Find the line of code of the source file where collection point it is.

    Returns:
        int: returns the line of code of the source file where collection point it is.

    """
    try:
        collection_point_file_path = inspect.stack()[2].filename
        collection_point_line_of_code = inspect.stack()[2].lineno
        is_temporal_file_path = _is_temporal_path(collection_point_file_path)
        if is_temporal_file_path:
            collection_point_line_of_code = UNKNOWN_LINE_OF_CODE
        return collection_point_line_of_code

    except Exception:
        return UNKNOWN_LINE_OF_CODE


def _is_temporal_path(path: str) -> bool:
    temporal_directory_path = tempfile.gettempdir()
    is_temporal_path = path.startswith(temporal_directory_path)
    return is_temporal_path


def _get_ipynb_file_path_collection() -> list[str]:
    current_working_directory_path = os.getcwd()
    cwd_file_name_collection = os.listdir(current_working_directory_path)
    ipynb_file_path_collection = []
    for file_name in cwd_file_name_collection:
        is_ipynb_file = file_name.endswith(DOT_IPYNB_EXTENSION)
        if is_ipynb_file:
            file_path = os.path.join(current_working_directory_path, file_name)
            ipynb_file_path_collection.append(file_path)

    return ipynb_file_path_collection



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import numpy as np
import pandas as pd
import pandera as pa

from snowflake.hypothesis_snowpark.constants import (
    PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES,
)
from snowflake.snowpark.types import (
    BooleanType,
    ByteType,
    DateType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
    TimestampTimeZone,
    TimestampType,
)
from snowflake.snowpark.types import (
    DataType as SnowparkDataType,
)


pandera_dtype_to_snowpark_dtype_dict = {
    np.dtypes.Int8DType: ByteType(),
    pd.Int8Dtype: ByteType(),
    np.dtypes.Int16DType: ShortType(),
    pd.Int16Dtype: ShortType(),
    np.dtypes.Int32DType: IntegerType(),
    pd.Int32Dtype: IntegerType(),
    np.dtypes.Int64DType: LongType(),
    pd.Int64Dtype: LongType(),
    np.dtypes.Float32DType: FloatType(),
    pd.Float32Dtype: FloatType(),
    np.dtypes.Float64DType: DoubleType(),
    pd.Float64Dtype: DoubleType(),
    np.dtypes.StrDType: StringType(),
    pd.StringDtype: StringType(),
    np.dtypes.BoolDType: BooleanType(),
    pd.BooleanDtype: BooleanType(),
    pd.DatetimeTZDtype: TimestampType(TimestampTimeZone.TZ),
    np.dtypes.DateTime64DType: TimestampType(TimestampTimeZone.NTZ),
    pa.engines.pandas_engine.Date: DateType(),
}


def pandera_dtype_to_snowpark_dtype(pandera_dtype: pa.DataType) -> SnowparkDataType:
    """Convert a Pandera data type to the equivalent Snowpark data type.

    Args:
        pandera_dtype: The Pandera data type to convert.

    Raises:
        TypeError: If the Pandera data type is not supported.

    Returns:
        The equivalent Snowpark data type.

    """
    snowpark_dtype = pandera_dtype_to_snowpark_dtype_dict.get(
        type(pandera_dtype)
    ) or pandera_dtype_to_snowpark_dtype_dict.get(type(pandera_dtype.type))

    if not snowpark_dtype:
        raise TypeError(f"Unsupported Pandera data type: {pandera_dtype}")
    return snowpark_dtype


def pyspark_dtype_to_snowpark_dtype(pyspark_dtype: str) -> SnowparkDataType:
    """Convert a PySpark data type to the equivalent Snowpark data type.

    Args:
        pyspark_dtype: The PySpark data type to convert.

    Raises:
        ValueError: If the PySpark data type is not supported.

    Returns:
        The equivalent Snowpark data type.

    """
    snowpark_type = PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES.get(pyspark_dtype)
    if not snowpark_type:
        raise TypeError(f"Unsupported PySpark data type: {pyspark_dtype}")
    return snowpark_type



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    ARRAY_COLUMN_TYPE,
    COLUMN_ALLOW_NULL_KEY,
    COLUMN_KEY_TYPE_KEY,
    COLUMN_METADATA_KEY,
    COLUMN_VALUE_TYPE_KEY,
    FIELD_METADATA_KEY,
    FIELDS_KEY,
    KEY_TYPE_KEY,
    MAP_COLUMN_TYPE,
    STRUCT_COLUMN_TYPE,
    VALUE_CONTAINS_NULL_KEY,
    VALUE_TYPE_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class EmptyColumnCollector(ColumnCollectorBase):

    """Class for collect an empty column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init EmptyColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        super().__init__(clm_name, struct_field, clm_values)

    def get_custom_data(self) -> dict[str, any]:
        custom_data = {}
        data_type_dict = self.struct_field.dataType.jsonValue()
        if self.type == ARRAY_COLUMN_TYPE:
            custom_data[COLUMN_VALUE_TYPE_KEY] = data_type_dict[VALUE_TYPE_KEY]
            custom_data[COLUMN_ALLOW_NULL_KEY] = data_type_dict[VALUE_CONTAINS_NULL_KEY]

        elif self.type == MAP_COLUMN_TYPE:
            custom_data[COLUMN_KEY_TYPE_KEY] = data_type_dict[KEY_TYPE_KEY]
            custom_data[COLUMN_VALUE_TYPE_KEY] = data_type_dict[VALUE_TYPE_KEY]
            custom_data[COLUMN_ALLOW_NULL_KEY] = data_type_dict[VALUE_CONTAINS_NULL_KEY]

        elif self.type == STRUCT_COLUMN_TYPE:
            field_metadata_collection = []
            struct_field_json = self.struct_field.dataType.jsonValue()
            for field in struct_field_json[FIELDS_KEY]:
                del field[FIELD_METADATA_KEY]
                field_metadata_collection.append(field)
            custom_data[COLUMN_METADATA_KEY] = field_metadata_collection

        else:
            custom_data = {}

        return custom_data



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import copy
import datetime
import json

from typing import Optional, Union

import pandera as pa

from dateutil.parser import parse
from hypothesis.strategies import DrawFn, SearchStrategy, composite

from snowflake.hypothesis_snowpark.checks.date_check import (
    dates_in_range,  # noqa: F401 Import required to register the custom checks
)
from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_COLUMNS_KEY,
    CUSTOM_DATA_FORMAT_KEY,
    CUSTOM_DATA_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_TYPE_KEY,
    PANDERA_IN_RANGE_CHECK,
    PANDERA_INCLUDE_MAX_KEY,
    PANDERA_INCLUDE_MIN_KEY,
    PANDERA_MAX_VALUE_KEY,
    PANDERA_MIN_VALUE_KEY,
    PANDERA_SCHEMA_KEY,
    PYSPARK_ARRAY_TYPE,
    PYSPARK_BINARY_TYPE,
    PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES,
)
from snowflake.hypothesis_snowpark.custom_strategies import update_pandas_df_strategy
from snowflake.hypothesis_snowpark.strategies_utils import (
    apply_custom_null_values,
    generate_snowpark_dataframe,
    generate_snowpark_schema,
    load_json_schema,
)
from snowflake.hypothesis_snowpark.telemetry.telemetry import report_telemetry
from snowflake.snowpark import DataFrame, Session


@report_telemetry(params_list=["schema"])
def dataframe_strategy(
    schema: Union[str, pa.DataFrameSchema], session: Session, size: Optional[int] = None
) -> SearchStrategy[DataFrame]:
    """Create a Hypothesis strategy for generating Snowpark DataFrames based on a Pandera JSON schema.

    Args:
        schema: The Pandera schema. This can be a path to a JSON schema file or a Pandera
                DataFrameSchema object.
        session: The Snowpark session to use for creating the DataFrames.
        size: The number of rows to generate. If not specified, the strategy will generate an arbitrary number of rows.

    Examples:
        Generate a Snowpark DataFrame from a JSON schema file:

        >>> from hypothesis import given
        >>> from snowflake.hypothesis_snowpark import dataframe_strategy
        >>> from snowflake.snowpark import DataFrame, Session
            <BLANKLINE>
        >>> @given(
        ...     df=dataframe_strategy(
        ...         schema="path/to/schema.json",
        ...         session=Session.builder.getOrCreate(),
        ...         size=10,
        ...     )
        ... )
        >>> def test_my_function(df: DataFrame):
        ...     ...

        Generate a Snowpark DataFrame from a Pandera DataFrameSchema object:

        >>> import pandera as pa
        >>> from hypothesis import given
        >>> from snowflake.hypothesis_snowpark import dataframe_strategy
        >>> from snowflake.snowpark import DataFrame, Session
            <BLANKLINE>
        >>> @given(
        ...    df=dataframe_strategy(
        ...        schema=pa.DataFrameSchema(
        ...            {
        ...                "A": pa.Column(pa.Int, checks=pa.Check.in_range(0, 10)),
        ...                "B": pa.Column(pa.Bool),
        ...            }
        ...        ),
        ...        session=Session.builder.getOrCreate(),
        ...        size=10,
        ...    )
        ... )
        >>> def test_my_function(df: DataFrame):
        ...     ...

    Returns:
        A Hypothesis strategy that generates Snowpark DataFrames.

    """
    if not session:
        raise ValueError("Session cannot be None.")

    if isinstance(schema, pa.DataFrameSchema):
        return _dataframe_strategy_from_object_schema(schema, session, size)

    if isinstance(schema, str) and schema.endswith(".json"):
        return _dataframe_strategy_from_json_schema(schema, session, size)

    raise ValueError(
        "Schema must be a path to a JSON schema file or a Pandera DataFrameSchema object."
    )


def _dataframe_strategy_from_object_schema(
    schema: pa.DataFrameSchema, session: Session, size: Optional[int] = None
) -> SearchStrategy[DataFrame]:
    """Create a Hypothesis strategy for generating Snowpark DataFrames based on a Pandera DataFrameSchema object.

    Args:
        schema: The Pandera DataFrameSchema object.
        session: The Snowpark session to use for creating the DataFrames.
        size: The number of rows to generate.

    Returns:
        A Hypothesis strategy that generates Snowpark DataFrames.

    """
    original_schema = copy.deepcopy(schema)

    for column in schema.columns.values():
        if isinstance(column.dtype, pa.engines.pandas_engine.Date):
            # Data generation for date type is currently unsupported by Pandera.
            # As a workaround, we can change the data type to pa.DateTime to
            # avoid an exception and let Snowpark handle the conversion to Date.
            column.dtype = pa.DateTime

    @composite
    def _dataframe_strategy(draw: DrawFn) -> DataFrame:
        pandas_strategy = schema.strategy(size=size)
        pandas_df = draw(pandas_strategy)
        snowpark_schema = generate_snowpark_schema(original_schema)
        snowpark_df = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
        return snowpark_df

    return _dataframe_strategy()


def _dataframe_strategy_from_json_schema(
    schema: str, session: Session, size: Optional[int] = None
) -> SearchStrategy[DataFrame]:
    """Create a Hypothesis strategy for generating Snowpark DataFrames based on a JSON schema.

    Args:
        schema: The path to the JSON schema file.
        session: The Snowpark session to use for creating the DataFrames.
        size: The number of rows to generate.

    Returns:
        A Hypothesis strategy that generates Snowpark DataFrames.

    """
    json_schema_dict = load_json_schema(schema)
    pandera_schema = json_schema_dict.get(PANDERA_SCHEMA_KEY)
    custom_data = json_schema_dict.get(CUSTOM_DATA_KEY)

    if not (pandera_schema and custom_data):
        raise ValueError(
            f"Invalid JSON schema. The JSON schema must contain '{PANDERA_SCHEMA_KEY}' and '{CUSTOM_DATA_KEY}' keys."
        )

    custom_data_columns = custom_data.get(CUSTOM_DATA_COLUMNS_KEY, [])
    not_supported_columns, columns_with_custom_strategy = [], []

    for column in custom_data_columns:
        dtype = column.get(CUSTOM_DATA_TYPE_KEY)
        if dtype not in PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES:
            not_supported_columns.append(column)
        elif dtype in (PYSPARK_ARRAY_TYPE, PYSPARK_BINARY_TYPE):
            columns_with_custom_strategy.append(column)

    if not_supported_columns:
        raise ValueError(
            f"The following data types are not supported by the Snowpark DataFrame strategy: "
            f"{[column.get(CUSTOM_DATA_TYPE_KEY) for column in not_supported_columns]}"
        )

    df_schema = pa.DataFrameSchema.from_json(json.dumps(pandera_schema))
    df_schema = _process_dataframe_schema(df_schema, custom_data)

    @composite
    def _dataframe_strategy(draw: DrawFn) -> DataFrame:
        pandas_strategy = df_schema.strategy(size=size)
        pandas_df = draw(pandas_strategy)
        pandas_df = draw(
            update_pandas_df_strategy(pandas_df, columns_with_custom_strategy)
        )
        pandas_df = apply_custom_null_values(pandas_df, custom_data)
        snowpark_schema = generate_snowpark_schema(df_schema, custom_data)
        snowpark_df = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
        return snowpark_df

    return _dataframe_strategy()


def _process_dataframe_schema(
    df_schema: pa.DataFrameSchema, custom_data: dict
) -> pa.DataFrameSchema:
    df_schema_copy = copy.copy(df_schema)

    for column_name, column_obj in df_schema_copy.columns.items():
        if type(column_obj.dtype) is pa.engines.pandas_engine.Date:
            # Data generation for date type is currently unsupported by Pandera. As a workaround, we can change the data
            # type to pa.DateTime to avoid an exception and manually generate the dates.
            column_obj.dtype = pa.DateTime

            in_range_check = next(
                (
                    check
                    for check in column_obj.checks
                    if check.name == PANDERA_IN_RANGE_CHECK
                ),
                None,
            )

            if in_range_check is None:
                # Generate the values as DateTime and let Snowpark handle the conversion to Date.
                continue

            min_value = in_range_check.statistics.get(PANDERA_MIN_VALUE_KEY)
            max_value = in_range_check.statistics.get(PANDERA_MAX_VALUE_KEY)
            include_min = in_range_check.statistics.get(PANDERA_INCLUDE_MIN_KEY, True)
            include_max = in_range_check.statistics.get(PANDERA_INCLUDE_MAX_KEY, True)
            date_format = next(
                (
                    column.get(CUSTOM_DATA_FORMAT_KEY)
                    for column in custom_data.get(CUSTOM_DATA_COLUMNS_KEY, [])
                    if column.get(CUSTOM_DATA_NAME_KEY) == column_name
                ),
                None,
            )

            if date_format is not None:
                min_value_obj = datetime.datetime.strptime(
                    min_value, date_format
                ).date()
                max_value_obj = datetime.datetime.strptime(
                    max_value, date_format
                ).date()
            else:
                min_value_obj = parse(min_value).date()
                max_value_obj = parse(max_value).date()

            # Replace the previous checks with the new date range check.
            column_obj.checks = [
                pa.Check.dates_in_range(
                    min_value=min_value_obj,
                    max_value=max_value_obj,
                    include_min=include_min,
                    include_max=include_max,
                )
            ]

    return df_schema_copy



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from statistics import mean

from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_IS_UNIQUE_SIZE_KEY,
    COLUMN_MAX_SIZE_KEY,
    COLUMN_MEAN_SIZE_KEY,
    COLUMN_MIN_SIZE_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class BinaryColumnCollector(ColumnCollectorBase):

    """Class for collect a binary type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init BinaryColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        super().__init__(clm_name, struct_field, clm_values)
        self._binary_size_collection = self._compute_binary_size_collection()

    def get_custom_data(self) -> dict[str, any]:
        max_size_value = max(self._binary_size_collection)
        min_size_value = min(self._binary_size_collection)
        mean_size_value = mean(self._binary_size_collection)
        all_elements_have_same_size = max_size_value == min_size_value

        custom_data_dict = {
            COLUMN_MAX_SIZE_KEY: max_size_value,
            COLUMN_MIN_SIZE_KEY: min_size_value,
            COLUMN_MEAN_SIZE_KEY: mean_size_value,
            COLUMN_IS_UNIQUE_SIZE_KEY: all_elements_have_same_size,
        }

        return custom_data_dict

    def _compute_binary_size_collection(self) -> list[int]:
        size_collection = []
        for binary in self.values:
            if binary is None:
                continue

            length = len(binary)
            size_collection.append(length)

        return size_collection



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_COUNT_KEY,
    COLUMN_METADATA_KEY,
    COLUMN_ROWS_NOT_NULL_COUNT_KEY,
    COLUMN_ROWS_NULL_COUNT_KEY,
    FIELD_METADATA_KEY,
    FIELDS_KEY,
    NAME_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class StructColumnCollector(ColumnCollectorBase):

    """Class for collect a struct type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init StructColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        super().__init__(clm_name, struct_field, clm_values)

    def get_custom_data(self) -> dict[str, any]:
        metadata = self._compute_struct_metadata()

        custom_data_dict = {
            COLUMN_METADATA_KEY: metadata,
        }

        return custom_data_dict

    def _compute_struct_metadata(self) -> list[dict[str, any]]:
        field_metadata_collection = []
        struct_field_json = self.struct_field.dataType.jsonValue()
        for field in struct_field_json[FIELDS_KEY]:
            del field[FIELD_METADATA_KEY]
            clm_name = field[NAME_KEY]
            rows_count_dict = self._compute_rows_count_by_column(clm_name)
            struct_field_custom_data = dict(field | rows_count_dict)
            field_metadata_collection.append(struct_field_custom_data)

        return field_metadata_collection

    def _compute_rows_count_by_column(self, clm_name: str) -> dict[str, int]:
        rows_count = 0
        rows_not_null_count = 0
        rows_null_count = 0
        for row in self.values:
            rows_count += 1
            if row is None:
                rows_null_count += 1
                continue

            row_clm_value = row[clm_name]
            if row_clm_value is None:
                rows_null_count += 1
            else:
                rows_not_null_count += 1

        rows_count_dict = {
            COLUMN_COUNT_KEY: rows_count,
            COLUMN_ROWS_NOT_NULL_COUNT_KEY: rows_not_null_count,
            COLUMN_ROWS_NULL_COUNT_KEY: rows_null_count,
        }

        return rows_count_dict



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

# Wrapper around pandera which logs to snowflake
from typing import Any, Optional, Union

from pandas import DataFrame as PandasDataFrame
from pandera import Check, DataFrameSchema
from pandera_report import DataFrameValidator

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.checkpoint_logger import CheckpointLogger
from snowflake.snowpark_checkpoints.utils.constant import (
    FAIL_STATUS,
    PASS_STATUS,
    CheckpointMode,
)
from snowflake.snowpark_checkpoints.utils.extra_config import is_checkpoint_enabled
from snowflake.snowpark_checkpoints.utils.telemetry import STATUS_KEY, report_telemetry
from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _add_custom_checks,
    _compare_data,
    _generate_schema,
    _process_sampling,
    _skip_checks_on_schema,
    _update_validation_result,
    _validate_checkpoint_name,
)


def validate_dataframe_checkpoint(
    df: SnowparkDataFrame,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    mode: Optional[CheckpointMode] = CheckpointMode.SCHEMA,
    custom_checks: Optional[dict[Any, Any]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Union[tuple[bool, PandasDataFrame], None]:
    """Validate a Snowpark DataFrame against a specified checkpoint.

    Args:
        df (SnowparkDataFrame): The DataFrame to validate.
        checkpoint_name (str): The name of the checkpoint to validate against.
        job_context (SnowparkJobContext, optional): The job context for the validation. Required for PARQUET mode.
        mode (CheckpointMode): The mode of validation (e.g., SCHEMA, PARQUET). Defaults to SCHEMA.
        custom_checks (Optional[dict[Any, Any]], optional): Custom checks to apply during validation.
        skip_checks (Optional[dict[Any, Any]], optional): Checks to skip during validation.
        sample_frac (Optional[float], optional): Fraction of the DataFrame to sample for validation. Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample for validation.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy to use for sampling.
            Defaults to RANDOM_SAMPLE.
        output_path (Optional[str], optional): The output path for the validation results.

    Returns:
        Union[tuple[bool, PandasDataFrame], None]: A tuple containing a boolean indicating success
        and a Pandas DataFrame with validation results, or None if validation is not applicable.

    Raises:
        ValueError: If an invalid validation mode is provided or if job_context is None for PARQUET mode.

    """
    _validate_checkpoint_name(checkpoint_name)

    if mode == CheckpointMode.SCHEMA:
        return _check_dataframe_schema_file(
            df,
            checkpoint_name,
            job_context,
            custom_checks,
            skip_checks,
            sample_frac,
            sample_number,
            sampling_strategy,
            output_path,
        )
    elif mode == CheckpointMode.DATAFRAME:
        if job_context is None:
            raise ValueError(
                "Connectionless mode is not supported for Parquet validation"
            )
        _compare_data(df, job_context, checkpoint_name, output_path)
    else:
        raise ValueError(
            """Invalid validation mode.
            Please use for schema validation use a 1 or for a full data validation use a 2 for schema validation."""
        )


def _check_dataframe_schema_file(
    df: SnowparkDataFrame,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    custom_checks: Optional[dict[Any, Any]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> tuple[bool, PandasDataFrame]:
    """Generate and checks the schema for a given DataFrame based on a checkpoint name.

    Args:
        df (SnowparkDataFrame): The DataFrame to be validated.
        checkpoint_name (str): The name of the checkpoint to retrieve the schema.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        custom_checks (dict[Any, Any], optional): Custom checks to be added to the schema.
            Defaults to None.
        skip_checks (dict[Any, Any], optional): Checks to be skipped.
            Defaults to None.
        sample_frac (float, optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (int, optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (SamplingStrategy, optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (str, optional): The output path for the validation results.

    Raises:
        SchemaValidationError: If the DataFrame fails schema validation.

    Returns:
        tuple[bool, PanderaDataFrame]: A tuple containing the validity flag and the Pandera DataFrame.

    """
    if df is None:
        raise ValueError("DataFrame is required")

    if checkpoint_name is None:
        raise ValueError("Checkpoint name is required")

    schema = _generate_schema(checkpoint_name, output_path)

    return check_dataframe_schema(
        df,
        schema,
        checkpoint_name,
        job_context,
        custom_checks,
        skip_checks,
        sample_frac,
        sample_number,
        sampling_strategy,
        output_path,
    )


def check_dataframe_schema(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    job_context: Optional[SnowparkJobContext] = None,
    custom_checks: Optional[dict[str, list[Check]]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> Union[tuple[bool, PandasDataFrame], None]:
    """Validate a DataFrame against a given Pandera schema using sampling techniques.

    Args:
        df (SnowparkDataFrame): The DataFrame to be validated.
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (str, optional): The name of the checkpoint to retrieve the schema.
            Defaults to None.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        custom_checks (dict[Any, Any], optional): Custom checks to be added to the schema.
            Defaults to None.
        skip_checks (dict[Any, Any], optional): Checks to be skipped.
            Defaults to None.
        sample_frac (float, optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (int, optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (SamplingStrategy, optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        output_path (str, optional): The output path for the validation results.

    Raises:
        SchemaValidationError: If the DataFrame fails schema validation.

    Returns:
        Union[tuple[bool, PandasDataFrame]|None]: A tuple containing the validity flag and the Pandas DataFrame.
        If the validation for that checkpoint is disabled it returns None.

    """
    _validate_checkpoint_name(checkpoint_name)

    if df is None:
        raise ValueError("DataFrame is required")

    if pandera_schema is None:
        raise ValueError("Schema is required")

    if is_checkpoint_enabled(checkpoint_name):
        return _check_dataframe_schema(
            df,
            pandera_schema,
            checkpoint_name,
            job_context,
            custom_checks,
            skip_checks,
            sample_frac,
            sample_number,
            sampling_strategy,
            output_path,
        )


@report_telemetry(
    params_list=["pandera_schema"],
    return_indexes=[(STATUS_KEY, 0)],
    multiple_return=True,
)
def _check_dataframe_schema(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    job_context: SnowparkJobContext = None,
    custom_checks: Optional[dict[str, list[Check]]] = None,
    skip_checks: Optional[dict[Any, Any]] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    output_path: Optional[str] = None,
) -> tuple[bool, PandasDataFrame]:
    _skip_checks_on_schema(pandera_schema, skip_checks)

    _add_custom_checks(pandera_schema, custom_checks)

    pandera_schema_upper, sample_df = _process_sampling(
        df, pandera_schema, job_context, sample_frac, sample_number, sampling_strategy
    )

    # Raises SchemaError on validation issues
    validation_result_status = PASS_STATUS
    try:
        validator = DataFrameValidator()
        validation_result = validator.validate(
            pandera_schema_upper, sample_df, validity_flag=True
        )

        if job_context is not None:
            job_context.mark_pass(checkpoint_name)

        return validation_result
    except Exception as pandera_ex:
        validation_result_status = FAIL_STATUS
        raise SchemaValidationError(
            "Snowpark output schema validation error",
            job_context,
            checkpoint_name,
            pandera_ex,
        ) from pandera_ex
    finally:
        _update_validation_result(
            checkpoint_name, validation_result_status, output_path
        )


@report_telemetry(params_list=["pandera_schema"])
def check_output_schema(
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    job_context: Optional[SnowparkJobContext] = None,
    output_path: Optional[str] = None,
):
    """Decorate to validate the schema of the output of a Snowpark function.

    Args:
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (Optional[str], optional): The name of the checkpoint to retrieve the schema.
        sample_frac (Optional[float], optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        output_path (Optional[str], optional): The output path for the validation results.

    """

    def check_output_with_decorator(snowpark_fn):
        """Decorate to validate the schema of the output of a Snowpark function.

        Args:
            snowpark_fn (function): The Snowpark function to validate.

        Returns:
            function: The decorated function.

        """
        _checkpoint_name = checkpoint_name
        if checkpoint_name is None:
            _checkpoint_name = snowpark_fn.__name__
        _validate_checkpoint_name(_checkpoint_name)

        def wrapper(*args, **kwargs):
            """Wrapp a function to validate the schema of the output of a Snowpark function.

            Args:
                *args: The arguments to the Snowpark function.
                **kwargs: The keyword arguments to the Snowpark function.

            Returns:
                Any: The result of the Snowpark function.

            """
            # Run the sampled data in snowpark
            snowpark_results = snowpark_fn(*args, **kwargs)
            sampler = SamplingAdapter(
                job_context, sample_frac, sample_number, sampling_strategy
            )
            sampler.process_args([snowpark_results])
            pandas_sample_args = sampler.get_sampled_pandas_args()

            # Raises SchemaError on validation issues
            validation_result_status = PASS_STATUS
            try:
                validator = DataFrameValidator()
                validation_result = validator.validate(
                    pandera_schema, pandas_sample_args[0]
                )

                if job_context is not None:
                    job_context.mark_pass(_checkpoint_name)

                logger = CheckpointLogger().get_logger()
                logger.info(
                    f"Checkpoint {_checkpoint_name} validation result:\n{validation_result}"
                )

            except Exception as pandera_ex:
                validation_result_status = FAIL_STATUS
                raise SchemaValidationError(
                    "Snowpark output schema validation error",
                    job_context,
                    _checkpoint_name,
                    pandera_ex,
                ) from pandera_ex
            finally:
                _update_validation_result(
                    _checkpoint_name, validation_result_status, output_path
                )
            return snowpark_results

        return wrapper

    return check_output_with_decorator


@report_telemetry(params_list=["pandera_schema"])
def check_input_schema(
    pandera_schema: DataFrameSchema,
    checkpoint_name: str,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    job_context: Optional[SnowparkJobContext] = None,
    output_path: Optional[str] = None,
):
    """Decorate factory for validating input DataFrame schemas before function execution.

    Args:
        pandera_schema (DataFrameSchema): The Pandera schema to validate against.
        checkpoint_name (Optional[str], optional): The name of the checkpoint to retrieve the schema.
        sample_frac (Optional[float], optional): Fraction of data to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): Number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): Strategy for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        job_context (SnowparkJobContext, optional): Context for job-related operations.
            Defaults to None.
        output_path (Optional[str], optional): The output path for the validation results.


    """

    def check_input_with_decorator(snowpark_fn):
        """Decorate that validates input schemas for the decorated function.

        Args:
            snowpark_fn (Callable): The function to be decorated with input schema validation.

        Raises:
            SchemaValidationError: If input data fails schema validation.

        Returns:
            Callable: A wrapper function that performs schema validation before executing the original function.

        """
        _checkpoint_name = checkpoint_name
        if checkpoint_name is None:
            _checkpoint_name = snowpark_fn.__name__
        _validate_checkpoint_name(_checkpoint_name)

        def wrapper(*args, **kwargs):
            """Wrapp a function to validate the schema of the input of a Snowpark function.

            Raises:
                SchemaValidationError: If any input DataFrame fails schema validation.

            Returns:
                Any: The result of the original function after input validation.

            """
            # Run the sampled data in snowpark
            sampler = SamplingAdapter(
                job_context, sample_frac, sample_number, sampling_strategy
            )
            sampler.process_args(args)
            pandas_sample_args = sampler.get_sampled_pandas_args()

            # Raises SchemaError on validation issues
            for arg in pandas_sample_args:
                validation_result_status = PASS_STATUS
                if isinstance(arg, PandasDataFrame):
                    try:
                        validator = DataFrameValidator()
                        validation_result = validator.validate(
                            pandera_schema,
                            arg,
                        )

                        if job_context is not None:
                            job_context.mark_pass(
                                _checkpoint_name,
                            )

                        logger = CheckpointLogger().get_logger()
                        logger.info(
                            f"Checkpoint {checkpoint_name} validation result:\n{validation_result}"
                        )

                    except Exception as pandera_ex:
                        validation_result_status = FAIL_STATUS
                        raise SchemaValidationError(
                            "Snowpark output schema validation error",
                            job_context,
                            _checkpoint_name,
                            pandera_ex,
                        ) from pandera_ex
                    finally:
                        _update_validation_result(
                            _checkpoint_name, validation_result_status, output_path
                        )
            return snowpark_fn(*args, **kwargs)

        return wrapper

    return check_input_with_decorator



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

__all__ = [
    "ArrayColumnCollector",
    "BinaryColumnCollector",
    "BooleanColumnCollector",
    "DateColumnCollector",
    "DayTimeIntervalColumnCollector",
    "DecimalColumnCollector",
    "EmptyColumnCollector",
    "MapColumnCollector",
    "NumericColumnCollector",
    "NullColumnCollector",
    "StringColumnCollector",
    "StructColumnCollector",
    "TimestampColumnCollector",
    "TimestampNTZColumnCollector",
]

from snowflake.snowpark_checkpoints_collector.column_collection.model.array_column_collector import (
    ArrayColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.binary_column_collector import (
    BinaryColumnCollector,
)

from snowflake.snowpark_checkpoints_collector.column_collection.model.boolean_column_collector import (
    BooleanColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.date_column_collector import (
    DateColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.day_time_interval_column_collector import (
    DayTimeIntervalColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.decimal_column_collector import (
    DecimalColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.empty_column_collector import (
    EmptyColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.map_column_collector import (
    MapColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.null_column_collector import (
    NullColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.numeric_column_collector import (
    NumericColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.string_column_collector import (
    StringColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.struct_column_collector import (
    StructColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.timestamp_column_collector import (
    TimestampColumnCollector,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.timestamp_ntz_column_collector import (
    TimestampNTZColumnCollector,
)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from snowflake.snowpark_checkpoints.utils.constant import (
    BINARY_TYPE,
    BOOLEAN_TYPE,
    BYTE_TYPE,
    DATE_TYPE,
    DECIMAL_TYPE,
    DOUBLE_TYPE,
    FLOAT_TYPE,
    INTEGER_TYPE,
    LONG_TYPE,
    SHORT_TYPE,
    STRING_TYPE,
    TIMESTAMP_NTZ_TYPE,
    TIMESTAMP_TYPE,
)


NumericTypes = [
    BYTE_TYPE,
    SHORT_TYPE,
    INTEGER_TYPE,
    LONG_TYPE,
    FLOAT_TYPE,
    DOUBLE_TYPE,
    DECIMAL_TYPE,
]

StringTypes = [STRING_TYPE]

BinaryTypes = [BINARY_TYPE]

BooleanTypes = [BOOLEAN_TYPE]

DateTypes = [DATE_TYPE, TIMESTAMP_TYPE, TIMESTAMP_NTZ_TYPE]

SupportedTypes = [
    BYTE_TYPE,
    SHORT_TYPE,
    INTEGER_TYPE,
    LONG_TYPE,
    FLOAT_TYPE,
    DOUBLE_TYPE,
    DECIMAL_TYPE,
    STRING_TYPE,
    BINARY_TYPE,
    BOOLEAN_TYPE,
    DATE_TYPE,
    TIMESTAMP_TYPE,
    TIMESTAMP_NTZ_TYPE,
]



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import os

from typing import Optional

from snowflake.snowpark_checkpoints.utils.constant import (
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
    VALIDATION_RESULTS_JSON_FILE_NAME,
)
from snowflake.snowpark_checkpoints.validation_results import (
    ValidationResult,
    ValidationResults,
)


class ValidationResultsMetadata:

    """ValidationResultsMetadata is a class that manages the loading, storing, and updating of validation results.

    Attributes:
        validation_results (list): A list to store validation results.
        validation_results_file (str): The path to the validation results file.

    Methods:
        __init__(path: Optional[str] = None):
            Initializes the PipelineResultMetadata instance and loads validation results from a JSON file
            if a path is provided.
        _load(path: Optional[str] = None):
            Loads validation results from a JSON file. If no path is provided, the current working directory is used.
        add_validation_result(validation_result: dict):
            Adds a validation result to the pipeline result list.
        save():
            Saves the validation results to a JSON file in the current working directory.

    """

    def __init__(self, path: Optional[str] = None):
        self._load(path)

    def _load(self, path: Optional[str] = None):
        """Load validation results from a JSON file.

        Args:
            path (Optional[str]): The directory path where the validation results file is located.
                                  If not provided, the current working directory is used.

        Raises:
            Exception: If there is an error reading the validation results file.

        """
        dir_path = path if path else os.getcwd()

        self.validation_results_file = os.path.join(
            dir_path,
            SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
            VALIDATION_RESULTS_JSON_FILE_NAME,
        )

        self.validation_results = ValidationResults(results=[])

        if os.path.exists(self.validation_results_file):
            with open(self.validation_results_file) as file:
                try:
                    validation_result_json = file.read()
                    self.validation_results = ValidationResults.model_validate_json(
                        validation_result_json
                    )
                except Exception as e:
                    raise Exception(
                        f"Error reading validation results file: {self.validation_results_file} \n {e}"
                    ) from None

    def add_validation_result(self, validation_result: ValidationResult):
        """Add a validation result to the pipeline result list.

        Args:
            checkpoint_name (str): The name of the checkpoint.
            validation_result (dict): The validation result to be added.

        """
        self.validation_results.results.append(validation_result)

    def save(self):
        """Save the validation results to a JSON file.

        The file is saved in the current working directory with a predefined
        file name. The validation results are serialized using the
        ValidationResultEncoder.

        Raises:
            OSError: If the file cannot be opened or written to.

        """
        if not os.path.exists(SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME):
            os.makedirs(SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME)

        with open(self.validation_results_file, "w") as output_file:
            output_file.write(self.validation_results.model_dump_json())



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from typing import Optional

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class Checkpoint(BaseModel):

    """Checkpoint model.

    Args:
        pydantic.BaseModel (pydantic.BaseModel): pydantic BaseModel

    """

    name: str
    mode: int = 1
    function: Optional[str] = None
    df: Optional[str] = None
    sample: Optional[float] = None
    file: Optional[str] = None
    location: int = -1
    enabled: bool = True


class Pipeline(BaseModel):

    """Pipeline model.

    Args:
        pydantic.BaseModel (pydantic.BaseModel): pydantic BaseModel

    """

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    entry_point: str
    checkpoints: list[Checkpoint]


class Checkpoints(BaseModel):

    """Checkpoints model.

    Args:
        pydantic.BaseModel (pydantic.BaseModel): pydantic BaseModel

    Returns:
        Checkpoints: An instance of the Checkpoints class

    """

    type: str
    pipelines: list[Pipeline]

    # this dictionary holds the unpacked checkpoints from the different pipelines.
    _checkpoints = {}

    def _build_checkpoints_dict(self):
        for pipeline in self.pipelines:
            for checkpoint in pipeline.checkpoints:
                self._checkpoints[checkpoint.name] = checkpoint

    def get_check_point(self, checkpoint_name: str) -> Checkpoint:
        """Get a checkpoint by its name.

        Args:
            checkpoint_name (str): The name of the checkpoint.

        Returns:
            Checkpoint: The checkpoint object if found, otherwise a new Checkpoint object
            with the name set to the checkpoint_id.

        """
        if not self._checkpoints:
            self._build_checkpoints_dict()

        checkpoint = self._checkpoints.get(checkpoint_name)
        if len(self._checkpoints) == 0:
            checkpoint = Checkpoint(name=checkpoint_name, enabled=True)
        elif checkpoint is None:
            checkpoint = Checkpoint(name=checkpoint_name, enabled=False)
        return checkpoint

    def add_checkpoint(self, checkpoint: Checkpoint) -> None:
        """Add a checkpoint to the checkpoints' dictionary.

        Args:.
            checkpoint (Checkpoint): The checkpoint object to add

        """
        self._checkpoints[checkpoint.name] = checkpoint



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import os

from typing import Optional

from snowflake.snowpark_checkpoints_collector.collection_common import (
    SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR,
    CheckpointMode,
)


# noinspection DuplicatedCode
def _get_checkpoint_contract_file_path() -> str:
    return os.environ.get(SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR, os.getcwd())


# noinspection DuplicatedCode
def _get_metadata():
    try:
        from snowflake.snowpark_checkpoints_configuration.checkpoint_metadata import (
            CheckpointMetadata,
        )

        path = _get_checkpoint_contract_file_path()
        metadata = CheckpointMetadata(path)
        return True, metadata

    except ImportError:
        return False, None


def is_checkpoint_enabled(checkpoint_name: str) -> bool:
    """Check if a checkpoint is enabled.

    Args:
        checkpoint_name (str): The name of the checkpoint.

    Returns:
        bool: True if the checkpoint is enabled, False otherwise.

    """
    enabled, metadata = _get_metadata()
    if enabled:
        config = metadata.get_checkpoint(checkpoint_name)
        return config.enabled
    else:
        return True


def get_checkpoint_sample(
    checkpoint_name: str, sample: Optional[float] = None
) -> float:
    """Get the checkpoint sample.

        Following this order first, the sample passed as argument, second, the sample from the checkpoint configuration,
        third, the default sample value 1.0.

    Args:
        checkpoint_name (str): The name of the checkpoint.
        sample (float, optional): The value passed to the function.

    Returns:
        float: returns the sample for that specific checkpoint.

    """
    default_sample = 1.0

    enabled, metadata = _get_metadata()
    if enabled:
        config = metadata.get_checkpoint(checkpoint_name)
        default_sample = config.sample if config.sample is not None else default_sample

    return sample if sample is not None else default_sample


def get_checkpoint_mode(
    checkpoint_name: str, mode: Optional[CheckpointMode] = None
) -> CheckpointMode:
    """Get the checkpoint mode.

        Following this order first, the mode passed as argument, second, the mode from the checkpoint configuration,
        third, the default mode value 1.

    Args:
        checkpoint_name (str): The name of the checkpoint.
        mode (int, optional): The value passed to the function.

    Returns:
        int: returns the mode for that specific checkpoint.

    """
    default_mode = CheckpointMode.SCHEMA

    enabled, metadata = _get_metadata()
    if enabled:
        config = metadata.get_checkpoint(checkpoint_name)
        default_mode = config.mode if config.mode is not None else default_mode

    return mode if mode is not None else default_mode



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from typing import Final

from snowflake.snowpark.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DataType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
    TimestampTimeZone,
    TimestampType,
)


PANDERA_SCHEMA_KEY: Final[str] = "pandera_schema"
PANDERA_IN_RANGE_CHECK: Final[str] = "in_range"
PANDERA_MIN_VALUE_KEY: Final[str] = "min_value"
PANDERA_MAX_VALUE_KEY: Final[str] = "max_value"
PANDERA_INCLUDE_MIN_KEY: Final[str] = "include_min"
PANDERA_INCLUDE_MAX_KEY: Final[str] = "include_max"

CUSTOM_DATA_KEY: Final[str] = "custom_data"
CUSTOM_DATA_FORMAT_KEY: Final[str] = "format"
CUSTOM_DATA_NAME_KEY: Final[str] = "name"
CUSTOM_DATA_COLUMNS_KEY: Final[str] = "columns"
CUSTOM_DATA_TYPE_KEY: Final[str] = "type"
CUSTOM_DATA_ROWS_COUNT: Final[str] = "rows_count"
CUSTOM_DATA_ROWS_NULL_COUNT_KEY: Final[str] = "rows_null_count"
CUSTOM_DATA_MIN_SIZE_KEY: Final[str] = "min_size"
CUSTOM_DATA_MAX_SIZE_KEY: Final[str] = "max_size"
CUSTOM_DATA_VALUE_TYPE_KEY: Final[str] = "value_type"

PYSPARK_ARRAY_TYPE: Final[str] = "array"
PYSPARK_BINARY_TYPE: Final[str] = "binary"
PYSPARK_BOOLEAN_TYPE: Final[str] = "boolean"
PYSPARK_BYTE_TYPE: Final[str] = "byte"
PYSPARK_DATE_TYPE: Final[str] = "date"
PYSPARK_DECIMAL_TYPE: Final[str] = "decimal"
PYSPARK_DOUBLE_TYPE: Final[str] = "double"
PYSPARK_FLOAT_TYPE: Final[str] = "float"
PYSPARK_INTEGER_TYPE: Final[str] = "integer"
PYSPARK_LONG_TYPE: Final[str] = "long"
PYSPARK_SHORT_TYPE: Final[str] = "short"
PYSPARK_STRING_TYPE: Final[str] = "string"
PYSPARK_TIMESTAMP_TYPE: Final[str] = "timestamp"
PYSPARK_TIMESTAMP_NTZ_TYPE: Final[str] = "timestamp_ntz"

PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES: Final[dict[str, DataType]] = {
    PYSPARK_ARRAY_TYPE: ArrayType(),
    PYSPARK_BINARY_TYPE: BinaryType(),
    PYSPARK_BOOLEAN_TYPE: BooleanType(),
    PYSPARK_BYTE_TYPE: ByteType(),
    PYSPARK_DATE_TYPE: DateType(),
    PYSPARK_DECIMAL_TYPE: DecimalType(),
    PYSPARK_DOUBLE_TYPE: DoubleType(),
    PYSPARK_FLOAT_TYPE: FloatType(),
    PYSPARK_INTEGER_TYPE: IntegerType(),
    PYSPARK_LONG_TYPE: LongType(),
    PYSPARK_SHORT_TYPE: ShortType(),
    PYSPARK_STRING_TYPE: StringType(),
    PYSPARK_TIMESTAMP_TYPE: TimestampType(),
    PYSPARK_TIMESTAMP_NTZ_TYPE: TimestampType(TimestampTimeZone.NTZ),
}



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from typing import Optional

import pandas

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext


class SamplingStrategy:
    RANDOM_SAMPLE = 1
    LIMIT = 2


class SamplingError(Exception):
    pass


class SamplingAdapter:
    def __init__(
        self,
        job_context: Optional[SnowparkJobContext],
        sample_frac: Optional[float] = None,
        sample_number: Optional[int] = None,
        sampling_strategy: SamplingStrategy = SamplingStrategy.RANDOM_SAMPLE,
    ):
        self.pandas_sample_args = []
        self.job_context = job_context
        if sample_frac and not (0 <= sample_frac <= 1):
            raise ValueError(
                f"'sample_size' value {sample_frac} is out of range (0 <= sample_size <= 1)"
            )

        self.sample_frac = sample_frac
        self.sample_number = sample_number
        self.sampling_strategy = sampling_strategy

    def process_args(self, input_args):
        # create the intermediate pandas
        # data frame for the test data
        for arg in input_args:
            if isinstance(arg, SnowparkDataFrame):
                if arg.count() == 0:
                    raise SamplingError(
                        "Input DataFrame is empty. Cannot sample from an empty DataFrame."
                    )

                if self.sampling_strategy == SamplingStrategy.RANDOM_SAMPLE:
                    if self.sample_frac:
                        df_sample = arg.sample(frac=self.sample_frac).to_pandas()
                    else:
                        df_sample = arg.sample(n=self.sample_number).to_pandas()
                else:
                    df_sample = arg.limit(self.sample_number).to_pandas()

                self.pandas_sample_args.append(df_sample)
            else:
                self.pandas_sample_args.append(arg)

    def get_sampled_pandas_args(self):
        return self.pandas_sample_args

    def get_sampled_snowpark_args(self):
        if self.job_context is None:
            raise SamplingError("Need a job context to compare with Spark")
        snowpark_sample_args = []
        for arg in self.pandas_sample_args:
            if isinstance(arg, pandas.DataFrame):
                snowpark_df = self.job_context.snowpark_session.create_dataframe(arg)
                snowpark_sample_args.append(snowpark_df)
            else:
                snowpark_sample_args.append(arg)
        return snowpark_sample_args

    def get_sampled_spark_args(self):
        if self.job_context is None:
            raise SamplingError("Need a job context to compare with Spark")
        pyspark_sample_args = []
        for arg in self.pandas_sample_args:
            if isinstance(arg, pandas.DataFrame):
                pyspark_df = self.job_context.spark_session.createDataFrame(arg)
                pyspark_sample_args.append(pyspark_df)
            else:
                pyspark_sample_args.append(arg)
        return pyspark_sample_args



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

# Skip type
from enum import IntEnum
from typing import Final


class CheckpointMode(IntEnum):
    SCHEMA = 1
    DATAFRAME = 2


# Execution mode
SCHEMA_EXECUTION_MODE: Final[str] = "Schema"
DATAFRAME_EXECUTION_MODE: Final[str] = "Dataframe"


# File position on stack
STACK_POSITION_CHECKPOINT: Final[int] = 6

# Validation status
PASS_STATUS: Final[str] = "PASS"
FAIL_STATUS: Final[str] = "FAIL"

# Validation result keys
DEFAULT_KEY: Final[str] = "default"


# Skip type
SKIP_ALL: Final[str] = "skip_all"

# Supported types
BOOLEAN_TYPE: Final[str] = "boolean"
BINARY_TYPE: Final[str] = "binary"
BYTE_TYPE: Final[str] = "byte"
CHAR_TYPE: Final[str] = "char"
DATE_TYPE: Final[str] = "date"
DAYTIMEINTERVAL_TYPE: Final[str] = "daytimeinterval"
DECIMAL_TYPE: Final[str] = "decimal"
DOUBLE_TYPE: Final[str] = "double"
FLOAT_TYPE: Final[str] = "float"
INTEGER_TYPE: Final[str] = "integer"
LONG_TYPE: Final[str] = "long"
SHORT_TYPE: Final[str] = "short"
STRING_TYPE: Final[str] = "string"
TIMESTAMP_TYPE: Final[str] = "timestamp"
TIMESTAMP_NTZ_TYPE: Final[str] = "timestamp_ntz"
VARCHAR_TYPE: Final[str] = "varchar"

# Pandas data types
PANDAS_BOOLEAN_DTYPE: Final[str] = "bool"
PANDAS_DATETIME_DTYPE: Final[str] = "datetime64[ns]"
PANDAS_FLOAT_DTYPE: Final[str] = "float64"
PANDAS_INTEGER_DTYPE: Final[str] = "int64"
PANDAS_OBJECT_DTYPE: Final[str] = "object"
PANDAS_TIMEDELTA_DTYPE: Final[str] = "timedelta64[ns]"

# Schemas keys
COLUMNS_KEY: Final[str] = "columns"
COUNT_KEY: Final[str] = "rows_count"
DECIMAL_PRECISION_KEY: Final[str] = "decimal_precision"
FALSE_COUNT_KEY: Final[str] = "false_count"
FORMAT_KEY: Final[str] = "format"
NAME_KEY: Final[str] = "name"
MARGIN_ERROR_KEY: Final[str] = "margin_error"
MAX_KEY: Final[str] = "max"
MEAN_KEY: Final[str] = "mean"
MIN_KEY: Final[str] = "min"
NULL_COUNT_KEY: Final[str] = "rows_null_count"
NULLABLE_KEY: Final[str] = "nullable"
ROWS_NOT_NULL_COUNT_KEY: Final[str] = "rows_not_null_count"
TRUE_COUNT_KEY: Final[str] = "true_count"
TYPE_KEY: Final[str] = "type"
ROWS_COUNT_KEY: Final[str] = "rows_count"

DATAFRAME_CUSTOM_DATA_KEY: Final[str] = "custom_data"
DATAFRAME_PANDERA_SCHEMA_KEY: Final[str] = "pandera_schema"

# SQL Column names
TABLE_NAME_COL: Final[str] = "TABLE_NAME"
CREATED_COL: Final[str] = "CREATED"

# SQL Table names
INFORMATION_SCHEMA_TABLE_NAME: Final[str] = "INFORMATION_SCHEMA"
TABLES_TABLE_NAME: Final[str] = "TABLES"

# SQL Query
EXCEPT_HASH_AGG_QUERY: Final[
    str
] = "SELECT HASH_AGG(*) FROM IDENTIFIER(:1) EXCEPT SELECT HASH_AGG(*) FROM IDENTIFIER(:2)"

# Table checkpoints name
CHECKPOINT_TABLE_NAME_FORMAT: Final[str] = "{}_snowpark"

# Write mode
OVERWRITE_MODE: Final[str] = "overwrite"

# Validation modes
VALIDATION_MODE_KEY: Final[str] = "validation_mode"
PIPELINES_KEY: Final[str] = "pipelines"

# File name
CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME: Final[str] = "{}.json"
CHECKPOINTS_JSON_FILE_NAME: Final[str] = "checkpoints.json"
SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME: Final[str] = "snowpark-checkpoints-output"
CHECKPOINT_PARQUET_OUTPUT_FILE_FORMAT_NAME: Final[str] = "{}.parquet"
VALIDATION_RESULTS_JSON_FILE_NAME: Final[str] = "checkpoint_validation_results.json"

# Environment variables
SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR: Final[
    str
] = "SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH"



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    ARRAY_COLUMN_TYPE,
    BINARY_COLUMN_TYPE,
    BOOLEAN_COLUMN_TYPE,
    BYTE_COLUMN_TYPE,
    DATE_COLUMN_TYPE,
    DAYTIMEINTERVAL_COLUMN_TYPE,
    DECIMAL_COLUMN_TYPE,
    DOUBLE_COLUMN_TYPE,
    FLOAT_COLUMN_TYPE,
    INTEGER_COLUMN_TYPE,
    LONG_COLUMN_TYPE,
    MAP_COLUMN_TYPE,
    NULL_COLUMN_TYPE,
    SHORT_COLUMN_TYPE,
    STRING_COLUMN_TYPE,
    STRUCT_COLUMN_TYPE,
    TIMESTAMP_COLUMN_TYPE,
    TIMESTAMP_NTZ_COLUMN_TYPE,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model import (
    ArrayColumnCollector,
    BinaryColumnCollector,
    BooleanColumnCollector,
    DateColumnCollector,
    DayTimeIntervalColumnCollector,
    DecimalColumnCollector,
    EmptyColumnCollector,
    MapColumnCollector,
    NullColumnCollector,
    NumericColumnCollector,
    StringColumnCollector,
    StructColumnCollector,
    TimestampColumnCollector,
    TimestampNTZColumnCollector,
)


def collector_register(cls):
    """Decorate a class with the collection type mechanism.

    Args:
        cls: The class to decorate.

    Returns:
        The class to decorate.

    """
    cls._collectors = {}
    for method_name in dir(cls):
        method = getattr(cls, method_name)
        if hasattr(method, "_column_type"):
            col_type_collection = method._column_type
            for col_type in col_type_collection:
                cls._collectors[col_type] = method_name
    return cls


def column_register(*args):
    """Decorate a method to register it in the collection mechanism based on column type.

    Args:
        args: the column type to register.

    Returns:
        The wrapper.

    """

    def wrapper(func):
        has_arguments = len(args) > 0
        if has_arguments:
            func._column_type = args
        return func

    return wrapper


@collector_register
class ColumnCollectorManager:

    """Manage class for column collector based on type."""

    def collect_column(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        """Collect the data of the column based on the column type.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            values (pandas.Series): the column values as Pandas.Series.

        Returns:
            dict[str, any]: The data collected.

        """
        clm_type = struct_field.dataType.typeName()
        if clm_type not in self._collectors:
            return {}

        func_name = self._collectors[clm_type]
        func = getattr(self, func_name)
        data = func(clm_name, struct_field, values)
        return data

    @column_register(ARRAY_COLUMN_TYPE)
    def _collect_array_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = ArrayColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(BINARY_COLUMN_TYPE)
    def _collect_binary_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = BinaryColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(BOOLEAN_COLUMN_TYPE)
    def _collect_boolean_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = BooleanColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(DATE_COLUMN_TYPE)
    def _collect_date_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = DateColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(DAYTIMEINTERVAL_COLUMN_TYPE)
    def _collect_day_time_interval_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = DayTimeIntervalColumnCollector(
            clm_name, struct_field, values
        )
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(DECIMAL_COLUMN_TYPE)
    def _collect_decimal_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = DecimalColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(MAP_COLUMN_TYPE)
    def _collect_map_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = MapColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(NULL_COLUMN_TYPE)
    def _collect_null_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = NullColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(
        BYTE_COLUMN_TYPE,
        SHORT_COLUMN_TYPE,
        INTEGER_COLUMN_TYPE,
        LONG_COLUMN_TYPE,
        FLOAT_COLUMN_TYPE,
        DOUBLE_COLUMN_TYPE,
    )
    def _collect_numeric_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = NumericColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(STRING_COLUMN_TYPE)
    def _collect_string_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = StringColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(STRUCT_COLUMN_TYPE)
    def _collect_struct_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = StructColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(TIMESTAMP_COLUMN_TYPE)
    def _collect_timestamp_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = TimestampColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    @column_register(TIMESTAMP_NTZ_COLUMN_TYPE)
    def _collect_timestampntz_type_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        column_collector = TimestampNTZColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data

    def collect_empty_custom_data(
        self, clm_name: str, struct_field: StructField, values: Series
    ) -> dict[str, any]:
        """Collect the data of a empty column.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            values (pandas.Series): the column values as Pandas.Series.

        Returns:
            dict[str, any]: The data collected.

        """
        column_collector = EmptyColumnCollector(clm_name, struct_field, values)
        collected_data = column_collector.get_data()
        return collected_data



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_DECIMAL_PRECISION_KEY,
    COLUMN_MARGIN_ERROR_KEY,
    COLUMN_MAX_KEY,
    COLUMN_MEAN_KEY,
    COLUMN_MIN_KEY,
    INTEGER_TYPE_COLLECTION,
    get_decimal_token,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class NumericColumnCollector(ColumnCollectorBase):

    """Class for collect an empty column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.
        is_integer (boolean): a flag to indicate if values are integer or do not.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init NumericColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        super().__init__(clm_name, struct_field, clm_values)
        self.is_integer = self.type in INTEGER_TYPE_COLLECTION

    def get_custom_data(self) -> dict[str, any]:
        local_values = self.values.dropna()
        min_value = local_values.min().item()
        max_value = local_values.max().item()
        mean_value = local_values.mean().item()
        decimal_precision = self._compute_decimal_precision()
        margin_error = local_values.std().item()

        custom_data_dict = {
            COLUMN_MIN_KEY: min_value,
            COLUMN_MAX_KEY: max_value,
            COLUMN_MEAN_KEY: mean_value,
            COLUMN_DECIMAL_PRECISION_KEY: decimal_precision,
            COLUMN_MARGIN_ERROR_KEY: margin_error,
        }

        return custom_data_dict

    def _compute_decimal_precision(self) -> int:
        if self.is_integer:
            return 0

        decimal_part_index = 1
        decimal_token = get_decimal_token()
        max_decimal_digits_counted = 0

        for value in self.values.dropna():
            value_str = str(value)
            value_split_by_token = value_str.split(decimal_token)
            decimal_part = value_split_by_token[decimal_part_index]
            decimal_digits_counted = len(decimal_part)
            max_decimal_digits_counted = max(
                decimal_digits_counted, max_decimal_digits_counted
            )

        return max_decimal_digits_counted



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from statistics import mean

from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_ALLOW_NULL_KEY,
    COLUMN_IS_UNIQUE_SIZE_KEY,
    COLUMN_KEY_TYPE_KEY,
    COLUMN_MAX_SIZE_KEY,
    COLUMN_MEAN_SIZE_KEY,
    COLUMN_MIN_SIZE_KEY,
    COLUMN_NULL_VALUE_PROPORTION_KEY,
    COLUMN_VALUE_TYPE_KEY,
    KEY_TYPE_KEY,
    VALUE_CONTAINS_NULL_KEY,
    VALUE_TYPE_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class MapColumnCollector(ColumnCollectorBase):

    """Class for collect a map type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init MapColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        super().__init__(clm_name, struct_field, clm_values)
        self._map_size_collection = self._compute_map_size_collection()

    def get_custom_data(self) -> dict[str, any]:
        data_type_dict = self.struct_field.dataType.jsonValue()
        key_type = data_type_dict[KEY_TYPE_KEY]
        value_type = data_type_dict[VALUE_TYPE_KEY]
        allow_null = data_type_dict[VALUE_CONTAINS_NULL_KEY]
        null_value_proportion = (
            self._compute_null_value_proportion() if allow_null else 0.0
        )
        array_max_size = max(self._map_size_collection)
        array_min_size = min(self._map_size_collection)
        array_mean_size = mean(self._map_size_collection)
        all_array_have_same_size = array_max_size == array_min_size

        custom_data_dict = {
            COLUMN_KEY_TYPE_KEY: key_type,
            COLUMN_VALUE_TYPE_KEY: value_type,
            COLUMN_ALLOW_NULL_KEY: allow_null,
            COLUMN_NULL_VALUE_PROPORTION_KEY: null_value_proportion,
            COLUMN_MAX_SIZE_KEY: array_max_size,
            COLUMN_MIN_SIZE_KEY: array_min_size,
            COLUMN_MEAN_SIZE_KEY: array_mean_size,
            COLUMN_IS_UNIQUE_SIZE_KEY: all_array_have_same_size,
        }

        return custom_data_dict

    def _compute_map_size_collection(self) -> list[int]:
        size_collection = []
        for mapp in self.values:
            if mapp is None:
                continue

            length = len(mapp)
            size_collection.append(length)

        return size_collection

    def _compute_null_value_proportion(self) -> float:
        null_counter = 0
        for mapp in self.values:
            if mapp is None:
                continue

            values_collection = list(mapp.values())
            null_counter += values_collection.count(None)

        total_values = sum(self._map_size_collection)
        null_value_proportion = (null_counter / total_values) * 100
        return null_value_proportion



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from snowflake.snowpark_checkpoints.checkpoint import (
    check_dataframe_schema,
    check_output_schema,
    check_input_schema,
    validate_dataframe_checkpoint,
)
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.spark_migration import check_with_spark

__all__ = [
    "check_with_spark",
    "SnowparkJobContext",
    "check_dataframe_schema",
    "check_output_schema",
    "check_input_schema",
    "validate_dataframe_checkpoint",
]



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import os

from typing import Optional

from snowflake.snowpark_checkpoints.utils.constant import (
    SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR,
)


# noinspection DuplicatedCode
def _get_checkpoint_contract_file_path() -> str:
    return os.environ.get(SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR, os.getcwd())


# noinspection DuplicatedCode
def _get_metadata():
    try:
        from snowflake.snowpark_checkpoints_configuration.checkpoint_metadata import (
            CheckpointMetadata,
        )

        path = _get_checkpoint_contract_file_path()
        metadata = CheckpointMetadata(path)
        return True, metadata

    except ImportError:
        return False, None


def is_checkpoint_enabled(checkpoint_name: Optional[str] = None) -> bool:
    """Check if a checkpoint is enabled.

    Args:
        checkpoint_name (Optional[str], optional): The name of the checkpoint.

    Returns:
        bool: True if the checkpoint is enabled, False otherwise.

    """
    enabled, metadata = _get_metadata()
    if enabled and checkpoint_name is not None:
        config = metadata.get_checkpoint(checkpoint_name)
        return config.enabled
    else:
        return True


def get_checkpoint_file(checkpoint_name: str) -> Optional[str]:
    """Retrieve the configuration for a specified checkpoint.

    This function fetches the checkpoint configuration if metadata is enabled.
    It extracts the file name from the checkpoint metadata or
    from the call stack if not explicitly provided in the metadata.

    Args:
        checkpoint_name (str): The name of the checkpoint to retrieve the configuration for.

    Returns:
        Optional[dict]: A dictionary containing the file name,
                        or None if metadata is not enabled.

    """
    enabled, metadata = _get_metadata()
    if enabled:
        config = metadata.get_checkpoint(checkpoint_name)

        return config.file
    else:
        return None



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

__all__ = ["SnowConnection"]

from snowflake.snowpark_checkpoints_collector.snow_connection_model.snow_connection import (
    SnowConnection,
)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import glob
import os.path
import time

from pathlib import Path
from typing import Callable, Optional

from snowflake.snowpark import Session
from snowflake.snowpark_checkpoints_collector.collection_common import (
    DOT_PARQUET_EXTENSION,
)


STAGE_NAME = "CHECKPOINT_STAGE"
CREATE_STAGE_STATEMENT_FORMAT = "CREATE TEMP STAGE IF NOT EXISTS {}"
REMOVE_STAGE_FOLDER_STATEMENT_FORMAT = "REMOVE {}"
STAGE_PATH_FORMAT = "'@{}/{}'"
PUT_FILE_IN_STAGE_STATEMENT_FORMAT = "PUT '{}' {} AUTO_COMPRESS=FALSE"


class SnowConnection:

    """Class for manage the Snowpark Connection.

    Attributes:
        session (Snowpark.Session): the Snowpark session.

    """

    def __init__(self, session: Session = None) -> None:
        """Init SnowConnection.

        Args:
            session (Snowpark.Session): the Snowpark session.

        """
        self.session = session if session is not None else Session.builder.getOrCreate()
        self.stage_id = int(time.time())

    def create_snowflake_table_from_local_parquet(
        self,
        table_name: str,
        input_path: str,
        stage_path: Optional[str] = None,
    ) -> None:
        """Upload to parquet files from the input path and create a table.

        Args:
            table_name (str): the name of the table to be created.
            input_path (str): the input directory path.
            stage_path: (str, optional): the stage path.

        """
        input_path = (
            os.path.abspath(input_path)
            if not os.path.isabs(input_path)
            else str(Path(input_path).resolve())
        )
        folder = f"table_files_{int(time.time())}"
        stage_path = stage_path if stage_path else folder
        stage_name = f"{STAGE_NAME}_{self.stage_id}"
        stage_directory_path = STAGE_PATH_FORMAT.format(stage_name, stage_path)

        def is_parquet_file(file: str):
            return file.endswith(DOT_PARQUET_EXTENSION)

        try:
            self.create_tmp_stage(stage_name)
            self.load_files_to_stage(
                stage_name, stage_path, input_path, is_parquet_file
            )
            self.create_table_from_parquet(table_name, stage_directory_path)

        finally:
            self.session.sql(
                REMOVE_STAGE_FOLDER_STATEMENT_FORMAT.format(stage_directory_path)
            ).collect()

    def create_tmp_stage(self, stage_name: str) -> None:
        """Create a temp stage in Snowflake.

        Args:
            stage_name (str): the name of the stage.

        """
        create_stage_statement = CREATE_STAGE_STATEMENT_FORMAT.format(stage_name)
        self.session.sql(create_stage_statement).collect()

    def load_files_to_stage(
        self,
        stage_name: str,
        folder_name: str,
        input_path: str,
        filter_func: Callable = None,
    ) -> None:
        """Load files to a stage in Snowflake.

        Args:
            stage_name (str): the name of the stage.
            folder_name (str): the folder name.
            input_path (str): the input directory path.
            filter_func (Callable): the filter function to apply to the files.

        """
        input_path = (
            os.path.abspath(input_path)
            if not os.path.isabs(input_path)
            else str(Path(input_path).resolve())
        )

        def filter_files(name: str):
            return os.path.isfile(name) and (filter_func(name) if filter_func else True)

        target_dir = os.path.join(input_path, "**", "*")
        files_collection = glob.glob(target_dir, recursive=True)

        files = [file for file in files_collection if filter_files(file)]

        if len(files) == 0:
            raise Exception(f"No files were found in the input directory: {input_path}")

        for file in files:
            # if file is relative path, convert to absolute path
            # if absolute path, then try to resolve as some Win32 paths are  not in LPN.
            file_full_path = (
                str(os.path.abspath(file))
                if not os.path.isabs(file)
                else str(Path(file).resolve())
            )
            # Snowflake required URI format for input in the put.
            normalize_file_path = Path(file_full_path).as_uri()
            new_file_path = file_full_path.replace(input_path, folder_name)
            # as Posix to convert Windows dir to posix
            new_file_path = Path(new_file_path).as_posix()
            stage_file_path = STAGE_PATH_FORMAT.format(stage_name, new_file_path)
            put_statement = PUT_FILE_IN_STAGE_STATEMENT_FORMAT.format(
                normalize_file_path, stage_file_path
            )
            self.session.sql(put_statement).collect()

    def create_table_from_parquet(
        self, table_name: str, stage_directory_path: str
    ) -> None:
        """Create a table from a parquet file in Snowflake.

        Args:
            table_name (str): the name of the table.
            stage_directory_path (str): the stage directory path.

        Raise:
            Exception: No parquet files were found in the stage

        """
        files = self.session.sql(f"LIST {stage_directory_path}").collect()
        if len(files) == 0:
            raise Exception("No parquet files were found in the stage.")
        dataframe = self.session.read.parquet(path=stage_directory_path)
        dataframe.write.save_as_table(table_name=table_name, mode="overwrite")



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#


from pydantic import BaseModel


class ValidationResult(BaseModel):

    """ValidationResult represents the result of a validation checkpoint.

    Attributes:
        result (str): The result of the validation.
        timestamp (datetime): The timestamp when the validation was performed.
        file (str): The file where the validation checkpoint is located.
        line_of_code (int): The line number in the file where the validation checkpoint is located.
        checkpoint_name (str): The name of the validation checkpoint.

    """

    result: str
    timestamp: str
    file: str
    line_of_code: int
    checkpoint_name: str


class ValidationResults(BaseModel):

    """ValidationResults is a model that holds a list of validation results.

    Attributes:
        results (list[ValidationResult]): A list of validation results.

    """

    results: list[ValidationResult]



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from snowflake.snowpark_checkpoints_configuration.checkpoint_metadata import (
    CheckpointMetadata,
)

__all__ = [
    "CheckpointMetadata",
]



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

__all__ = [
    "ColumnCollectorManager",
]

from snowflake.snowpark_checkpoints_collector.column_collection.column_collector_manager import (
    ColumnCollectorManager,
)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import inspect
import json
import os
import re

from datetime import datetime
from typing import Any, Optional

import numpy as np

from pandera import Check, DataFrameSchema

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark_checkpoints.errors import SchemaValidationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.checkpoint_logger import CheckpointLogger
from snowflake.snowpark_checkpoints.utils.constant import (
    CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME,
    CHECKPOINT_TABLE_NAME_FORMAT,
    COLUMNS_KEY,
    DATAFRAME_CUSTOM_DATA_KEY,
    DATAFRAME_EXECUTION_MODE,
    DATAFRAME_PANDERA_SCHEMA_KEY,
    DECIMAL_PRECISION_KEY,
    DEFAULT_KEY,
    EXCEPT_HASH_AGG_QUERY,
    FAIL_STATUS,
    FALSE_COUNT_KEY,
    MARGIN_ERROR_KEY,
    MEAN_KEY,
    NAME_KEY,
    NULL_COUNT_KEY,
    NULLABLE_KEY,
    PASS_STATUS,
    ROWS_COUNT_KEY,
    SKIP_ALL,
    SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME,
    TRUE_COUNT_KEY,
    TYPE_KEY,
)
from snowflake.snowpark_checkpoints.utils.extra_config import (
    get_checkpoint_file,
)
from snowflake.snowpark_checkpoints.utils.supported_types import (
    BooleanTypes,
    NumericTypes,
)
from snowflake.snowpark_checkpoints.validation_result_metadata import (
    ValidationResultsMetadata,
)
from snowflake.snowpark_checkpoints.validation_results import ValidationResult


def _validate_checkpoint_name(checkpoint_name: str) -> None:
    """Validate the given checkpoint name.

    Args:
        checkpoint_name (str): The name of the checkpoint to validate.

    Raises:
        ValueError: If the checkpoint name is invalid.
        Checkpoint names must only contain alphanumeric characters and underscores.

    """
    if not _is_valid_checkpoint_name(checkpoint_name):
        raise ValueError(
            f"Invalid checkpoint name: {checkpoint_name}. "
            "Checkpoint names must only contain alphanumeric characters and underscores."
        )


def _is_valid_checkpoint_name(checkpoint_name: str) -> bool:
    """Check if the provided checkpoint name is valid.

    A valid checkpoint name must:
    - Start with a letter (a-z, A-Z) or an underscore (_)
    - Be followed by any combination of letters, digits (0-9), underscores (_), or dollar signs ($)

    Args:
        checkpoint_name (str): The checkpoint name to validate.

    Returns:
        bool: True if the checkpoint name is valid, False otherwise.

    """
    regex = r"^[a-zA-Z_][a-zA-Z0-9_$]*$"
    return bool(re.match(regex, checkpoint_name))


def _process_sampling(
    df: SnowparkDataFrame,
    pandera_schema: DataFrameSchema,
    job_context: Optional[SnowparkJobContext] = None,
    sample_frac: Optional[float] = 1.0,
    sample_number: Optional[int] = None,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
):
    """Process a Snowpark DataFrame by sampling it according to the specified parameters.

    Adjusts the column casing of the provided Pandera schema to uppercase.

    Args:
        df (SnowparkDataFrame): The Snowpark DataFrame to be sampled.
        pandera_schema (DataFrameSchema): The Pandera schema to validate the DataFrame.
        job_context (SnowparkJobContext, optional): The job context for the sampling operation.
            Defaults to None.
        sample_frac (Optional[float], optional): The fraction of rows to sample.
            Defaults to 0.1.
        sample_number (Optional[int], optional): The number of rows to sample.
            Defaults to None.
        sampling_strategy (Optional[SamplingStrategy], optional): The strategy to use for sampling.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.

    Returns:
        Tuple[DataFrameSchema, pd.DataFrame]: A tuple containing the adjusted Pandera schema with uppercase column names
        and the sampled pandas DataFrame.

    """
    sampler = SamplingAdapter(
        job_context, sample_frac, sample_number, sampling_strategy
    )
    sampler.process_args([df])

    # fix up the column casing
    pandera_schema_upper = pandera_schema
    new_columns: dict[Any, Any] = {}

    for col in pandera_schema.columns:
        new_columns[col.upper()] = pandera_schema.columns[col]

    pandera_schema_upper = pandera_schema_upper.remove_columns(pandera_schema.columns)
    pandera_schema_upper = pandera_schema_upper.add_columns(new_columns)

    sample_df = sampler.get_sampled_pandas_args()[0]
    sample_df.index = np.ones(sample_df.count().iloc[0])

    return pandera_schema_upper, sample_df


def _add_numeric_checks(
    schema: DataFrameSchema, col: str, additional_check: dict[str, Any]
):
    """Add numeric checks to a specified column in a DataFrameSchema.

    Args:
        schema (DataFrameSchema): The schema to which the checks will be added.
        col (str): The name of the column to which the checks will be applied.
        additional_check (dict[str, Any]): A dictionary containing additional checks.
            - MEAN_KEY (str): The key for the mean value to check against.
            - MARGIN_ERROR_KEY (str): The key for the margin of error for the mean check.
            - DECIMAL_PRECISION_KEY (str): The key for the maximum decimal precision allowed.

    Returns:
        None

    """
    mean = additional_check.get(MEAN_KEY, 0)
    std = additional_check.get(MARGIN_ERROR_KEY, 0)

    def check_mean(series):
        series_mean = series.mean()
        return mean - std <= series_mean <= mean + std

    schema.columns[col].checks.append(
        Check(check_mean, element_wise=False, name="mean")
    )

    if DECIMAL_PRECISION_KEY in additional_check:
        schema.columns[col].checks.append(
            Check(
                lambda series: series.apply(
                    lambda x: len(str(x).split(".")[1]) if "." in str(x) else 0
                )
                <= additional_check[DECIMAL_PRECISION_KEY],
                name="decimal_precision",
            )
        )


def _add_boolean_checks(
    schema: DataFrameSchema, col: str, additional_check: dict[str, Any]
):
    """Add boolean checks to a specified column in a DataFrameSchema.

    Args:
        schema (DataFrameSchema): The schema to which the checks will be added.
        col (str): The name of the column to which the checks will be applied.
        additional_check (dict[str, Any]): A dictionary containing additional check parameters.
            - TRUE_COUNT_KEY (int): Expected count of True values in the column.
            - FALSE_COUNT_KEY (int): Expected count of False values in the column.
            - MARGIN_ERROR_KEY (int): Margin of error allowed for the counts.

    Returns:
        None

    """
    count_of_true = additional_check.get(TRUE_COUNT_KEY, 0)
    count_of_false = additional_check.get(FALSE_COUNT_KEY, 0)
    rows_count = additional_check.get(ROWS_COUNT_KEY, 0)
    std = additional_check.get(MARGIN_ERROR_KEY, 0)
    percentage_true = count_of_true / rows_count
    percentage_false = count_of_false / rows_count

    schema.columns[col].checks.extend(
        [
            Check(
                lambda series: (
                    percentage_true - std
                    <= series.value_counts().get(True, 0) / series.count()
                    if series.count() > 0
                    else 1 <= percentage_true + std
                ),
            ),
            Check(
                lambda series: (
                    percentage_false - std
                    <= series.value_counts().get(False, 0) / series.count()
                    if series.count() > 0
                    else 1 <= percentage_false + std
                ),
            ),
        ]
    )


def _add_null_checks(
    schema: DataFrameSchema, col: str, additional_check: dict[str, Any]
):
    """Add null checks to a specified column in a DataFrameSchema.

    Args:
        schema (DataFrameSchema): The schema to which the checks will be added.
        col (str): The name of the column to which the checks will be applied.
        additional_check (dict[str, Any]): A dictionary containing additional check parameters.
            - NULL_COUNT_KEY (int): Expected count of Null values in the column.
            - ROWS_COUNT_KEY (int): Total number of rows in the column.
            - MARGIN_ERROR_KEY (int): Margin of error allowed for the counts.

    Returns:
        None

    """
    count_of_null = additional_check.get(NULL_COUNT_KEY, 0)
    rows_count = additional_check.get(ROWS_COUNT_KEY, 0)
    std = additional_check.get(MARGIN_ERROR_KEY, 0)
    percentage_null = count_of_null / rows_count

    schema.columns[col].checks.extend(
        [
            Check(
                lambda series: (
                    percentage_null - std <= series.isnull().sum() / series.count()
                    if series.count() > 0
                    else 1 <= percentage_null + std
                ),
            ),
        ]
    )


def _generate_schema(
    checkpoint_name: str, output_path: Optional[str] = None
) -> DataFrameSchema:
    """Generate a DataFrameSchema based on the checkpoint name provided.

    This function reads a JSON file corresponding to the checkpoint name,
    extracts schema information, and constructs a DataFrameSchema object.
    It also adds custom checks for numeric and boolean types if specified
    in the JSON file.

    Args:
        checkpoint_name (str): The name of the checkpoint used to locate
                               the JSON file containing schema information.
        output_path (str): The path to the output directory.

        DataFrameSchema: A schema object representing the structure and
                         constraints of the DataFrame.
                         constraints of the DataFrame.

    """
    current_directory_path = output_path if output_path else os.getcwd()

    output_directory_path = os.path.join(
        current_directory_path, SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME
    )

    if not os.path.exists(output_directory_path):
        raise ValueError(
            """Output directory snowpark-checkpoints-output does not exist.
Please run the Snowpark checkpoint collector first."""
        )

    checkpoint_file_path = os.path.join(
        output_directory_path,
        CHECKPOINT_JSON_OUTPUT_FILE_FORMAT_NAME.format(checkpoint_name),
    )

    if not os.path.exists(checkpoint_file_path):
        raise ValueError(
            f"Checkpoint {checkpoint_name} JSON file not found. Please run the Snowpark checkpoint collector first."
        )

    with open(checkpoint_file_path) as custom_data_schema:
        custom_data_schema_json = json.load(custom_data_schema)

    if DATAFRAME_PANDERA_SCHEMA_KEY not in custom_data_schema_json:
        raise ValueError(
            f"Pandera schema not found in the JSON file for checkpoint: {checkpoint_name}"
        )

    schema_dict = custom_data_schema_json.get(DATAFRAME_PANDERA_SCHEMA_KEY)
    schema_dict_str = json.dumps(schema_dict)
    schema = DataFrameSchema.from_json(schema_dict_str)

    if DATAFRAME_CUSTOM_DATA_KEY not in custom_data_schema_json:
        return schema

    custom_data = custom_data_schema_json.get(DATAFRAME_CUSTOM_DATA_KEY)

    if COLUMNS_KEY not in custom_data:
        raise ValueError(
            f"Columns not found in the JSON file for checkpoint: {checkpoint_name}"
        )

    logger = CheckpointLogger().get_logger()

    for additional_check in custom_data.get(COLUMNS_KEY):

        type = additional_check.get(TYPE_KEY, None)
        name = additional_check.get(NAME_KEY, None)
        is_nullable = additional_check.get(NULLABLE_KEY, False)

        if name is None:
            raise ValueError(f"Column name not defined in the schema {checkpoint_name}")

        if type is None:
            raise ValueError(f"Type not defined for column {name}")

        if schema.columns.get(name) is None:
            logger.warning(f"Column {name} not found in schema")
            continue

        if type in NumericTypes:
            _add_numeric_checks(schema, name, additional_check)

        elif type in BooleanTypes:
            _add_boolean_checks(schema, name, additional_check)

        if is_nullable:
            _add_null_checks(schema, name, additional_check)

    return schema


def _skip_checks_on_schema(
    pandera_schema: DataFrameSchema,
    skip_checks: Optional[dict[str, list[str]]] = None,
) -> None:
    """Modify a Pandera DataFrameSchema to skip specified checks on certain columns.

    Args:
        pandera_schema (DataFrameSchema): The Pandera DataFrameSchema object to modify.
        skip_checks (Optional[dict[str, list[str]]]): A dictionary where keys are column names
                                                and values are lists of checks to skip for
                                                those columns. If the list is empty, all
                                                checks for the column will be removed.

    Returns:
        None

    """
    if not skip_checks:
        return

    for col, checks_to_skip in skip_checks.items():

        if col in pandera_schema.columns:

            if SKIP_ALL in checks_to_skip:
                pandera_schema.columns[col].checks = {}

            else:
                pandera_schema.columns[col].checks = [
                    check
                    for check in pandera_schema.columns[col].checks
                    if check.name not in checks_to_skip
                ]


def _add_custom_checks(
    schema: DataFrameSchema,
    custom_checks: Optional[dict[str, list[Check]]] = None,
):
    """Add custom checks to a Pandera DataFrameSchema.

    Args:
        schema (DataFrameSchema): The Pandera DataFrameSchema object to modify.
        custom_checks (Optional[dict[str, list[Check]]]): A dictionary where keys are column names
                                                and values are lists of checks to add for
                                                those columns.

    Returns:
        None

    """
    if not custom_checks:
        return

    for col, checks in custom_checks.items():
        if col in schema.columns:
            col_schema = schema.columns[col]
            col_schema.checks.extend(checks)
        else:
            raise ValueError(f"Column {col} not found in schema")


def _compare_data(
    df: SnowparkDataFrame,
    job_context: Optional[SnowparkJobContext],
    checkpoint_name: str,
    output_path: Optional[str] = None,
):
    """Compare the data in the provided Snowpark DataFrame with the data in a checkpoint table.

    This function writes the provided DataFrame to a table and compares it with an existing checkpoint table
    using a hash aggregation query. If there is a data mismatch, it marks the job context as failed and raises a
    SchemaValidationError. If the data matches, it marks the job context as passed.

    Args:
        df (SnowparkDataFrame): The Snowpark DataFrame to compare.
        job_context (SnowparkJobContext): The job context containing the Snowpark session and job state.
        checkpoint_name (str): The name of the checkpoint table to compare against.
        output_path (str): The path to the output directory.

    Raises:
        SchemaValidationError: If there is a data mismatch between the DataFrame and the checkpoint table.

    """
    new_table_name = CHECKPOINT_TABLE_NAME_FORMAT.format(checkpoint_name)

    df.write.save_as_table(table_name=new_table_name, mode="overwrite")

    expect_df = job_context.snowpark_session.sql(
        EXCEPT_HASH_AGG_QUERY, [checkpoint_name, new_table_name]
    )

    if expect_df.count() != 0:
        error_message = f"Data mismatch for checkpoint {checkpoint_name}"
        job_context.mark_fail(
            error_message,
            checkpoint_name,
            df,
            DATAFRAME_EXECUTION_MODE,
        )
        _update_validation_result(
            checkpoint_name,
            FAIL_STATUS,
            output_path,
        )
        raise SchemaValidationError(
            error_message,
            job_context,
            checkpoint_name,
            df,
        )
    else:
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
        job_context.mark_pass(checkpoint_name, DATAFRAME_EXECUTION_MODE)


def _find_frame_in(stack: list[inspect.FrameInfo]) -> tuple:
    """Find a specific frame in the provided stack trace.

    This function searches through the provided stack trace to find a frame that matches
    certain criteria. It looks for frames where the function name is "wrapper" or where
    the code context matches specific regular expressions.

    Args:
        stack (list[inspect.FrameInfo]): A list of frame information objects representing
                                         the current stack trace.

    Returns:
        tuple: A tuple containing the relative path of the file and the line number of the
               matched frame. If no frame is matched, it returns a default key and -1.

    """
    regex = (
        r"(?<!_check_dataframe_schema_file)"
        r"(?<!_check_dataframe_schema)"
        r"(validate_dataframe_checkpoint|check_dataframe_schema)"
    )

    first_frames = stack[:7]
    first_frames.reverse()

    for i, frame in enumerate(first_frames):
        if frame.function == "wrapper" and i - 1 >= 0:
            next_frame = first_frames[i - 1]
            return _get_relative_path(next_frame.filename), next_frame.lineno

        if len(frame.code_context) >= 0 and re.search(regex, frame.code_context[0]):
            return _get_relative_path(frame.filename), frame.lineno
    return DEFAULT_KEY, -1


def _get_relative_path(file_path: str) -> str:
    """Get the relative path of a file.

    Args:
        file_path (str): The path to the file.

    Returns:
        str: The relative path of the file.

    """
    current_directory = os.getcwd()
    return os.path.relpath(file_path, current_directory)


def _update_validation_result(
    checkpoint_name: str, validation_status: str, output_path: Optional[str] = None
) -> None:
    """Update the validation result file with the status of a given checkpoint.

    Args:
        checkpoint_name (str): The name of the checkpoint to update.
        validation_status (str): The validation status to record for the checkpoint.
        output_path (str): The path to the output directory.

    Returns:
        None

    """
    _file = get_checkpoint_file(checkpoint_name)

    stack = inspect.stack()

    _file_from_stack, _line_of_code = _find_frame_in(stack)

    pipeline_result_metadata = ValidationResultsMetadata(output_path)

    pipeline_result_metadata.add_validation_result(
        ValidationResult(
            timestamp=datetime.now().isoformat(),
            file=_file if _file else _file_from_stack,
            line_of_code=_line_of_code,
            checkpoint_name=checkpoint_name,
            result=validation_status,
        )
    )

    pipeline_result_metadata.save()



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import glob
import json
import os
import shutil

from typing import Optional

import pandera as pa

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.functions import col
from pyspark.sql.types import DoubleType as SparkDoubleType
from pyspark.sql.types import StringType as SparkStringType
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    CHECKPOINT_JSON_OUTPUT_FILE_NAME_FORMAT,
    COLUMNS_KEY,
    DATAFRAME_CUSTOM_DATA_KEY,
    DATAFRAME_PANDERA_SCHEMA_KEY,
    DECIMAL_COLUMN_TYPE,
    DOT_PARQUET_EXTENSION,
    NULL_COLUMN_TYPE,
    PANDAS_OBJECT_TYPE_COLLECTION,
    CheckpointMode,
)
from snowflake.snowpark_checkpoints_collector.collection_result.model import (
    CollectionPointResult,
    CollectionPointResultManager,
    CollectionResult,
)
from snowflake.snowpark_checkpoints_collector.column_collection import (
    ColumnCollectorManager,
)
from snowflake.snowpark_checkpoints_collector.column_pandera_checks import (
    PanderaColumnChecksManager,
)
from snowflake.snowpark_checkpoints_collector.snow_connection_model import (
    SnowConnection,
)
from snowflake.snowpark_checkpoints_collector.utils import file_utils
from snowflake.snowpark_checkpoints_collector.utils.extra_config import (
    get_checkpoint_mode,
    get_checkpoint_sample,
    is_checkpoint_enabled,
)
from snowflake.snowpark_checkpoints_collector.utils.telemetry import report_telemetry


def collect_dataframe_checkpoint(
    df: SparkDataFrame,
    checkpoint_name: str,
    sample: Optional[float] = None,
    mode: Optional[CheckpointMode] = None,
    output_path: Optional[str] = None,
) -> None:
    """Collect a DataFrame checkpoint.

    Args:
        df (SparkDataFrame): The input Spark DataFrame to collect.
        checkpoint_name (str): The name of the checkpoint.
        sample (float, optional): Fraction of DataFrame to sample for schema inference.
            Defaults to 1.0.
        mode (CheckpointMode): The mode to execution the collection.
            Defaults to CheckpointMode.Schema
        output_path (str, optional): The output path to save the checkpoint.
            Defaults to Current working Directory.

    Raises:
        Exception: Invalid mode value.

    """
    collection_point_file_path = file_utils.get_collection_point_source_file_path()
    collection_point_line_of_code = file_utils.get_collection_point_line_of_code()
    collection_point_result = CollectionPointResult(
        collection_point_file_path, collection_point_line_of_code, checkpoint_name
    )

    try:
        if is_checkpoint_enabled(checkpoint_name):

            _sample = get_checkpoint_sample(checkpoint_name, sample)

            if _is_empty_dataframe_without_schema(df):
                raise Exception(
                    "It is not possible to collect an empty DataFrame without schema"
                )

            _mode = get_checkpoint_mode(checkpoint_name, mode)

            if _mode == CheckpointMode.SCHEMA:
                column_type_dict = _get_spark_column_types(df)
                _collect_dataframe_checkpoint_mode_schema(
                    checkpoint_name, df, _sample, column_type_dict, output_path
                )

            elif _mode == CheckpointMode.DATAFRAME:
                snow_connection = SnowConnection()
                _collect_dataframe_checkpoint_mode_dataframe(
                    checkpoint_name, df, snow_connection, output_path
                )

            else:
                raise Exception("Invalid mode value.")

            collection_point_result.result = CollectionResult.PASS

    except Exception as err:
        collection_point_result.result = CollectionResult.FAIL
        error_message = str(err)
        raise Exception(error_message) from err

    finally:
        collection_point_result_manager = CollectionPointResultManager(output_path)
        collection_point_result_manager.add_result(collection_point_result)


@report_telemetry(params_list=["column_type_dict"])
def _collect_dataframe_checkpoint_mode_schema(
    checkpoint_name: str,
    df: SparkDataFrame,
    sample: float,
    column_type_dict: dict[str, any],
    output_path: Optional[str] = None,
) -> None:
    source_df = df.sample(sample)
    if source_df.isEmpty():
        source_df = df
    pandas_df = source_df.toPandas()
    is_empty_df_with_object_column = _is_empty_dataframe_with_object_column(df)
    pandera_infer_schema = (
        pa.infer_schema(pandas_df) if not is_empty_df_with_object_column else {}
    )

    column_name_collection = df.schema.names
    columns_to_remove_from_pandera_schema_collection = []
    column_custom_data_collection = []
    column_collector_manager = ColumnCollectorManager()
    column_pandera_checks_manager = PanderaColumnChecksManager()

    for column in column_name_collection:
        struct_field_column = column_type_dict[column]
        column_type = struct_field_column.dataType.typeName()

        is_empty_column = (
            len(pandas_df[column].dropna()) == 0 and column_type is not NULL_COLUMN_TYPE
        )
        is_column_to_remove_from_pandera_schema = (
            _is_column_to_remove_from_pandera_schema(column_type)
        )

        if is_column_to_remove_from_pandera_schema:
            columns_to_remove_from_pandera_schema_collection.append(column)

        if is_empty_column:
            custom_data = column_collector_manager.collect_empty_custom_data(
                column, struct_field_column, pandas_df[column]
            )
            column_custom_data_collection.append(custom_data)
            continue

        pandera_column = pandera_infer_schema.columns[column]
        pandera_column.checks = []
        column_pandera_checks_manager.add_checks_column(
            column, column_type, pandas_df, pandera_column
        )

        custom_data = column_collector_manager.collect_column(
            column, struct_field_column, pandas_df[column]
        )
        column_custom_data_collection.append(custom_data)

    pandera_infer_schema_dict = _get_pandera_infer_schema_as_dict(
        pandera_infer_schema,
        is_empty_df_with_object_column,
        columns_to_remove_from_pandera_schema_collection,
    )

    dataframe_custom_column_data = {COLUMNS_KEY: column_custom_data_collection}
    dataframe_schema_contract = {
        DATAFRAME_PANDERA_SCHEMA_KEY: pandera_infer_schema_dict,
        DATAFRAME_CUSTOM_DATA_KEY: dataframe_custom_column_data,
    }

    dataframe_schema_contract_json = json.dumps(dataframe_schema_contract)
    _generate_json_checkpoint_file(
        checkpoint_name, dataframe_schema_contract_json, output_path
    )


def _get_spark_column_types(df: SparkDataFrame) -> dict[str, StructField]:
    schema = df.schema
    column_type_collection = {}
    for field in schema.fields:
        column_name = field.name
        column_type_collection[column_name] = field
    return column_type_collection


def _is_empty_dataframe_without_schema(df: SparkDataFrame) -> bool:
    is_empty = df.isEmpty()
    has_schema = len(df.schema.fields) > 0
    return is_empty and not has_schema


def _is_empty_dataframe_with_object_column(df: SparkDataFrame):
    is_empty = df.isEmpty()
    if not is_empty:
        return False

    for field in df.schema.fields:
        if field.dataType.typeName() in PANDAS_OBJECT_TYPE_COLLECTION:
            return True

    return False


def _is_column_to_remove_from_pandera_schema(column_type) -> bool:
    is_decimal_type = column_type == DECIMAL_COLUMN_TYPE
    return is_decimal_type


def _get_pandera_infer_schema_as_dict(
    pandera_infer_schema, is_empty_df_with_string_column, columns_to_remove_collection
) -> dict[str, any]:
    if is_empty_df_with_string_column:
        return {}

    pandera_infer_schema_dict = json.loads(pandera_infer_schema.to_json())
    for column in columns_to_remove_collection:
        del pandera_infer_schema_dict[COLUMNS_KEY][column]

    return pandera_infer_schema_dict


def _generate_json_checkpoint_file(
    checkpoint_name, dataframe_schema_contract, output_path: Optional[str] = None
) -> None:

    checkpoint_file_name = CHECKPOINT_JSON_OUTPUT_FILE_NAME_FORMAT.format(
        checkpoint_name
    )
    output_directory_path = file_utils.get_output_directory_path(output_path)
    checkpoint_file_path = os.path.join(output_directory_path, checkpoint_file_name)
    with open(checkpoint_file_path, "w") as f:
        f.write(dataframe_schema_contract)


def _collect_dataframe_checkpoint_mode_dataframe(
    checkpoint_name: str,
    df: SparkDataFrame,
    snow_connection: SnowConnection,
    output_path: Optional[str] = None,
) -> None:
    output_path = file_utils.get_output_directory_path(output_path)
    parquet_directory = os.path.join(output_path, checkpoint_name)
    generate_parquet_for_spark_df(df, parquet_directory)
    _create_snowflake_table_from_parquet(
        checkpoint_name, parquet_directory, snow_connection
    )


def generate_parquet_for_spark_df(spark_df: SparkDataFrame, output_path: str) -> None:
    """Generate a parquet file from a Spark DataFrame.

    This function will  convert Float to Double to avoid precision problems.
    Spark parquet use IEEE 32-bit floating point values,
    while Snowflake uses IEEE 64-bit floating point values.

    Args:
        spark_df: dataframe to be saved as parquet
        output_path: path to save the parquet files.
    returns: None

    Raises:
        Exception: No parquet files were generated.

    """
    new_cols = [
        (
            col(c).cast(SparkStringType()).cast(SparkDoubleType()).alias(c)
            if t == "float"
            else col(c)
        )
        for (c, t) in spark_df.dtypes
    ]
    converted_df = spark_df.select(new_cols)

    if os.path.exists(output_path):
        shutil.rmtree(output_path)

    converted_df.write.parquet(output_path, mode="overwrite")

    target_dir = os.path.join(output_path, "**", f"*{DOT_PARQUET_EXTENSION}")
    files = glob.glob(target_dir, recursive=True)
    if len(files) == 0:
        raise Exception("No parquet files were generated.")


def _create_snowflake_table_from_parquet(
    table_name: str, input_path: str, snow_connection: SnowConnection
) -> None:
    snow_connection.create_snowflake_table_from_local_parquet(table_name, input_path)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import logging
import threading


class CheckpointLogger:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super().__new__(cls, *args, **kwargs)
                    cls._instance._initialize()
        return cls._instance

    def _initialize(self):
        # Create formatter
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        self.logger = logging.getLogger("CheckpointLogger")
        self.logger.setLevel(logging.INFO)

        # Create console handler and set level to debug
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        # Add formatter to ch
        ch.setFormatter(formatter)

        # Add ch to logger
        self.logger.addHandler(ch)

    def get_logger(self):
        return self.logger



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#


from functools import wraps
from typing import Callable


snowpark_strategies: dict[str, Callable] = {}


def register_strategy(dtype: str) -> Callable:
    """Register a strategy for a given data type.

    Args:
        dtype: The data type to register the strategy for.

    Returns:
        The decorated function.

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        snowpark_strategies[dtype] = func
        return wrapper

    return decorator



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from typing import Callable, Optional, TypeVar

import pandas as pd

from pyspark.sql import DataFrame as SparkDataFrame

from snowflake.snowpark import DataFrame as SnowparkDataFrame
from snowflake.snowpark.types import PandasDataFrame
from snowflake.snowpark_checkpoints.errors import SparkMigrationError
from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext
from snowflake.snowpark_checkpoints.snowpark_sampler import (
    SamplingAdapter,
    SamplingStrategy,
)
from snowflake.snowpark_checkpoints.utils.constant import FAIL_STATUS, PASS_STATUS
from snowflake.snowpark_checkpoints.utils.telemetry import STATUS_KEY, report_telemetry
from snowflake.snowpark_checkpoints.utils.utils_checks import (
    _update_validation_result,
    _validate_checkpoint_name,
)


fn = TypeVar("F", bound=Callable)


def check_with_spark(
    job_context: Optional[SnowparkJobContext],
    spark_function: fn,
    checkpoint_name: str,
    sample_number: Optional[int] = 100,
    sampling_strategy: Optional[SamplingStrategy] = SamplingStrategy.RANDOM_SAMPLE,
    check_dtypes: Optional[bool] = True,
    check_with_precision: Optional[float] = True,
    output_path: Optional[str] = None,
) -> Callable[[fn], fn]:
    """Validate function output with Spark instance.

    Will take the input snowpark dataframe of this function, sample data, convert
    it to a Spark dataframe and then execute `spark_function`. Subsequently
    the output of that function will be compared to the output of this function
    for the same sample of data.

    Args:
        job_context (SnowparkJobContext): The job context containing configuration and details for the validation.
        spark_function (fn): The equivalent PySpark function to compare against the Snowpark implementation.
        checkpoint_name (str): A name for the checkpoint. Defaults to None.
        sample_number (Optional[int], optional): The number of rows for validation. Defaults to 100.
        sampling_strategy (Optional[SamplingStrategy], optional): The strategy used for sampling data.
            Defaults to SamplingStrategy.RANDOM_SAMPLE.
        check_dtypes (Optional[bool], optional): Enable data type consistency checks between Snowpark and PySpark.
            Defaults to True.
        check_with_precision (Optional[float], optional): Precision value to control numerical comparison precision.
            Defaults to True.
        output_path (Optional[str], optional): The path to store the validation results. Defaults to None.

    Returns:
        Callable[[fn], fn]: A decorator that wraps the original Snowpark function with validation logic.

    """

    def check_with_spark_decorator(snowpark_fn):
        _checkpoint_name = checkpoint_name
        if checkpoint_name is None:
            _checkpoint_name = snowpark_fn.__name__
        _validate_checkpoint_name(_checkpoint_name)

        def wrapper(*args, **kwargs):
            sampler = SamplingAdapter(
                job_context,
                sample_number=sample_number,
                sampling_strategy=sampling_strategy,
            )
            sampler.process_args(args)
            snowpark_sample_args = sampler.get_sampled_snowpark_args()
            pyspark_sample_args = sampler.get_sampled_spark_args()
            # Run the sampled data in snowpark
            snowpark_test_results = snowpark_fn(*snowpark_sample_args, **kwargs)
            spark_test_results = spark_function(*pyspark_sample_args, **kwargs)
            result, exception = _assert_return(
                snowpark_test_results,
                spark_test_results,
                job_context,
                _checkpoint_name,
                output_path,
            )
            if not result:
                raise exception from None
            # Run the original function in snowpark
            return snowpark_fn(*args, **kwargs)

        return wrapper

    return check_with_spark_decorator


@report_telemetry(
    params_list=["snowpark_results", "spark_results"],
    return_indexes=[(STATUS_KEY, 0)],
    multiple_return=True,
)
def _assert_return(
    snowpark_results, spark_results, job_context, checkpoint_name, output_path=None
) -> tuple[bool, Optional[Exception]]:
    """Assert and validate the results from Snowpark and Spark transformations.

    Args:
        snowpark_results (Any): Results from the Snowpark transformation.
        spark_results (Any): Results from the Spark transformation to compare against.
        job_context (Any): Additional context about the job. Defaults to None.
        checkpoint_name (Any): Name of the checkpoint for logging. Defaults to None.
        output_path (Optional[str], optional): The path to store the validation results. Defaults to None.

    Raises:
        AssertionError: If the Snowpark and Spark results do not match.
        TypeError: If the results cannot be compared.

    """
    if isinstance(snowpark_results, SnowparkDataFrame) and isinstance(
        spark_results, SparkDataFrame
    ):
        cmp = compare_spark_snowpark_dfs(spark_results, snowpark_results)

        if not cmp.empty:
            exception_result = SparkMigrationError(
                "DataFrame difference:\n", job_context, checkpoint_name, cmp
            )
            return False, exception_result
        job_context.mark_pass(checkpoint_name)
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
        return True, None
    else:

        if snowpark_results != spark_results:
            exception_result = SparkMigrationError(
                "Return value difference:\n",
                job_context,
                checkpoint_name,
                f"{snowpark_results} != {spark_results}",
            )
            _update_validation_result(checkpoint_name, FAIL_STATUS, output_path)
            return False, exception_result
        job_context.mark_pass(checkpoint_name)
        _update_validation_result(checkpoint_name, PASS_STATUS, output_path)
        return True, None


def compare_spark_snowpark_dfs(
    spark_df: SparkDataFrame, snowpark_df: SnowparkDataFrame
) -> PandasDataFrame:
    """Compare two dataframes for equality.

    Args:
        spark_df (SparkDataFrame): The Spark dataframe to compare.
        snowpark_df (SnowparkDataFrame): The Snowpark dataframe to compare.

    Returns:Pandas DataFrame containing the differences between the two dataframes.

    """
    snowpark_df = snowpark_df.to_pandas()
    snowpark_df.columns = snowpark_df.columns.str.upper()
    spark_df = spark_df.toPandas()
    spark_df.columns = spark_df.columns.str.upper()
    spark_cols = set(spark_df.columns)
    snowpark_cols = set(snowpark_df.columns)
    cmp = pd.DataFrame([])
    left = spark_cols - snowpark_cols
    right = snowpark_cols - spark_cols
    if left != set():
        cmp = _compare_dfs(spark_df, snowpark_df, "spark", "snowpark")
    if right != set():
        right_cmp = _compare_dfs(snowpark_df, spark_df, "snowpark", "spark")
        cmp = right_cmp if cmp.empty else pd.concat([cmp, right_cmp], ignore_index=True)
    if left == set() and right == set():
        if spark_df.shape == snowpark_df.shape:
            cmp = spark_df.compare(snowpark_df, result_names=("spark", "snowpark"))
        else:
            cmp = spark_df.merge(snowpark_df, indicator=True, how="outer").loc[
                lambda x: x["_merge"] != "both"
            ]
            cmp = cmp.replace(
                {"left_only": "spark_only", "right_only": "snowpark_only"}
            )

    return cmp


def _compare_dfs(
    df_a: pd.DataFrame, df_b: pd.DataFrame, left_label: str, right_label: str
) -> PandasDataFrame:
    """Compare two dataframes for equality.

    Args:
        df_a (PandasDataFrame): The first dataframe to compare.
        df_b (PandasDataFrame): The second dataframe to compare.
        left_label (str): The label for the first dataframe.
        right_label (str): The label for the second dataframe.

    :return: Pandas DataFrame containing the differences between the two dataframes.

    """
    df_a["side"] = "a"
    df_b["side"] = "b"
    a_only = [col for col in df_a.columns if col not in df_b.columns] + ["side"]
    b_only = [col for col in df_b.columns if col not in df_a.columns] + ["side"]
    cmp = (
        df_a[a_only]
        .merge(df_b[b_only], indicator=True, how="left")
        .loc[lambda x: x["_merge"] != "both"]
    )
    cmp = cmp.replace(
        {"left_only": f"{left_label}_only", "right_only": f"{right_label}_only"}
    )
    cmp = cmp.drop(columns="side")
    return cmp



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import json
import os
import random

from contextlib import contextmanager
from typing import Optional, Union

import numpy as np
import pandas as pd
import pandera as pa

from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_COLUMNS_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_ROWS_COUNT,
    CUSTOM_DATA_ROWS_NULL_COUNT_KEY,
    CUSTOM_DATA_TYPE_KEY,
)
from snowflake.hypothesis_snowpark.datatype_mapper import (
    pandera_dtype_to_snowpark_dtype,
    pyspark_dtype_to_snowpark_dtype,
)
from snowflake.snowpark import DataFrame, Session
from snowflake.snowpark.types import StructField, StructType


def load_json_schema(json_schema: str) -> dict:
    """Load the JSON schema from a file.

    Args:
        json_schema: The path to the JSON file.

    Returns:
        The JSON schema as a dictionary.

    """
    if not os.path.isfile(json_schema):
        raise ValueError(f"Invalid JSON schema path: {json_schema}")

    try:
        with open(json_schema, encoding="utf-8") as file:
            return json.load(file)
    except (OSError, json.JSONDecodeError) as e:
        raise ValueError(f"Error reading JSON schema file: {e}") from None


def generate_snowpark_dataframe(
    pandas_df: pd.DataFrame,
    snowpark_schema: StructType,
    session: Session,
) -> DataFrame:
    """Generate a Snowpark DataFrame from a Pandas DataFrame.

    Args:
        pandas_df: The Pandas DataFrame to convert.
        snowpark_schema: The Snowpark schema to use for creating the DataFrame.
        session: The Snowpark session to use for creating the DataFrame.

    Returns:
        A Snowpark DataFrame.

    """
    pandas_df = pandas_df.replace({np.nan: None})

    # Snowpark ignores the schema argument if the data argument is a pandas dataframe. We need to convert the pandas
    # dataframe into a list of tuples to be able to specify a schema.
    data = list(pandas_df.itertuples(index=False, name=None))

    # In some scenarios, the create_dataframe function performs some random operations (e.g., generating temporary table
    # names or temporary stage names). Due to Hypothesis controls randomness during its tests, we need to temporarily
    # set a random seed to ensure that the names generated by Snowpark are unique to avoid conflicts.
    with temporary_random_seed():
        snowpark_df = session.create_dataframe(data=data, schema=snowpark_schema)
    return snowpark_df


def apply_custom_null_values(
    pandas_df: pd.DataFrame, custom_data: dict
) -> pd.DataFrame:
    """Apply null values to a Pandas DataFrame based on the custom data.

    This function modifies the DataFrame in place and returns a new DataFrame with null values applied.

    Args:
        pandas_df: The Pandas DataFrame to apply null values to.
        custom_data: The custom data associated with the Pandera schema.

    Returns:
        A Pandas DataFrame with null values applied.

    """
    null_proportions = {}
    for col in custom_data.get(CUSTOM_DATA_COLUMNS_KEY, []):
        rows_count = col.get(CUSTOM_DATA_ROWS_COUNT, 0)
        if rows_count > 0:
            null_proportions[col[CUSTOM_DATA_NAME_KEY]] = (
                col[CUSTOM_DATA_ROWS_NULL_COUNT_KEY] / rows_count
            )

    df_copy = pandas_df.copy()
    total_rows = len(df_copy)

    for column, target_null_proportion in null_proportions.items():
        if column not in df_copy.columns:
            continue

        required_nulls = round(total_rows * target_null_proportion)
        current_nulls = df_copy[column].isnull().sum()

        if current_nulls < required_nulls:
            additional_nulls = required_nulls - current_nulls
            non_null_positions = np.where(df_copy[column].notnull())[0]
            selected_positions = np.random.choice(
                non_null_positions, size=additional_nulls, replace=False
            )
            df_copy[column] = df_copy[column].astype(object)
            df_copy.iloc[selected_positions, df_copy.columns.get_loc(column)] = None

    return df_copy.replace({np.nan: None})


@contextmanager
def temporary_random_seed(seed: Union[int, float, str, bytes, bytearray, None] = None):
    """Set a temporary random seed within a context.

    This context manager saves the current state of the random number generator,
    sets a new seed, and then restores the original state when exiting the context.

    Args:
        seed: The seed to set for the random number generator. If not provided, seeds from current time or from an
              operating system specific randomness source if available.

    Yields:
        None: The context manager does not return anything.

    """
    current_state = random.getstate()
    random.seed(seed)
    try:
        yield
    finally:
        random.setstate(current_state)


def generate_snowpark_schema(
    pandera_schema: pa.DataFrameSchema, custom_data: Optional[dict] = None
) -> StructType:
    """Generate a Snowpark schema from a Pandera schema.

    Args:
        pandera_schema: The Pandera schema to convert.
        custom_data: The custom data associated with the Pandera schema.

    Returns:
        A Snowpark struct type.

    """
    struct_fields = []

    if custom_data is None:
        for column_name, column_data in pandera_schema.columns.items():
            snowpark_dtype = pandera_dtype_to_snowpark_dtype(column_data.dtype)
            struct_field = StructField(
                column_identifier=column_name,
                datatype=snowpark_dtype,
                nullable=column_data.nullable,
            )
            struct_fields.append(struct_field)
    else:
        for column_name, column_data in pandera_schema.columns.items():
            custom_data_column = next(
                (
                    column
                    for column in custom_data.get(CUSTOM_DATA_COLUMNS_KEY, [])
                    if column.get(CUSTOM_DATA_NAME_KEY) == column_name
                ),
                None,
            )

            if custom_data_column is None:
                raise ValueError(f"Column '{column_name}' is missing from custom_data")

            pyspark_type = custom_data_column.get(CUSTOM_DATA_TYPE_KEY)

            if pyspark_type is None:
                raise ValueError(
                    f"Type for column '{column_name}' is missing from custom_data"
                )

            snowpark_type = pyspark_dtype_to_snowpark_dtype(pyspark_type)

            struct_field = StructField(
                column_identifier=column_name,
                datatype=snowpark_type,
                nullable=column_data.nullable,
            )
            struct_fields.append(struct_field)

    return StructType(struct_fields)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import datetime as dt

from decimal import Decimal
from fractions import Fraction
from typing import Optional, Union

import hypothesis.strategies as st
import pandas as pd

from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_MAX_SIZE_KEY,
    CUSTOM_DATA_MIN_SIZE_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_TYPE_KEY,
    CUSTOM_DATA_VALUE_TYPE_KEY,
    PYSPARK_ARRAY_TYPE,
    PYSPARK_BINARY_TYPE,
    PYSPARK_BOOLEAN_TYPE,
    PYSPARK_BYTE_TYPE,
    PYSPARK_DATE_TYPE,
    PYSPARK_DECIMAL_TYPE,
    PYSPARK_DOUBLE_TYPE,
    PYSPARK_FLOAT_TYPE,
    PYSPARK_INTEGER_TYPE,
    PYSPARK_LONG_TYPE,
    PYSPARK_SHORT_TYPE,
    PYSPARK_STRING_TYPE,
)
from snowflake.hypothesis_snowpark.strategy_register import (
    register_strategy,
    snowpark_strategies,
)


@register_strategy(PYSPARK_ARRAY_TYPE)
@st.composite
def array_strategy(
    draw: st.DrawFn, dtype: str, min_size: int = 0, max_size: Optional[int] = None
) -> list:
    """Generate a list of values for a given data type.

    Args:
        draw: The Hypothesis draw function.
        dtype: The data type of the array.
        min_size: The minimum size of the array.
        max_size: The maximum size of the array.

    Returns:
        A list of values for the given data type.

    """
    if dtype not in snowpark_strategies:
        raise ValueError(f"Not implemented SearchStrategy for arrays of type '{dtype}'")

    strategy = snowpark_strategies[dtype]
    return draw(st.lists(strategy(), min_size=min_size, max_size=max_size))


@register_strategy(PYSPARK_BINARY_TYPE)
@st.composite
def binary_strategy(
    draw: st.DrawFn, min_size: int = 0, max_size: Optional[int] = None
) -> bytes:
    """Generate a binary data (bytes) with a given length.

    Args:
        draw: The Hypothesis draw function.
        min_size: The minimum length of the binary data.
        max_size: The maximum length of the binary data.

    Returns:
        A binary data (bytes) with a given length.

    """
    return draw(st.binary(min_size=min_size, max_size=max_size))


@register_strategy(PYSPARK_BOOLEAN_TYPE)
@st.composite
def boolean_strategy(draw: st.DrawFn) -> bool:
    """Generate a boolean value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A boolean value.

    """
    return draw(st.booleans())


@register_strategy(PYSPARK_BYTE_TYPE)
@st.composite
def byte_strategy(draw: st.DrawFn) -> int:
    """Generate a byte value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A byte value.

    """
    return draw(st.integers(min_value=-128, max_value=127))


@register_strategy(PYSPARK_DATE_TYPE)
@st.composite
def date_strategy(
    draw: st.DrawFn, min_value: dt.date = dt.date.min, max_value: dt.date = dt.date.max
) -> dt.date:
    """Generate a date value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum date value.
        max_value: The maximum date value.

    Returns:
        A date value within a given range.

    """
    return draw(st.dates(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_DECIMAL_TYPE)
@st.composite
def decimal_strategy(
    draw: st.DrawFn,
    min_value: Union[int, float, Fraction, Decimal, str, None] = None,
    max_value: Union[int, float, Fraction, Decimal, str, None] = None,
) -> Decimal:
    """Generate a decimal value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum value of the decimal.
        max_value: The maximum value of the decimal.

    Returns:
        A decimal value within a given range.

    """
    return draw(st.decimals(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_FLOAT_TYPE)
@register_strategy(PYSPARK_DOUBLE_TYPE)
@st.composite
def float_strategy(
    draw: st.DrawFn,
    min_value: Union[int, float, Fraction, Decimal, None] = None,
    max_value: Union[int, float, Fraction, Decimal, None] = None,
) -> float:
    """Generate a float value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum value of the float.
        max_value: The maximum value of the float.

    Returns:
        A float value within a given range.

    """
    return draw(st.floats(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_INTEGER_TYPE)
@st.composite
def integer_strategy(
    draw: st.DrawFn,
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
) -> int:
    """Generate an integer value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum value of the integer.
        max_value: The maximum value of the integer.

    Returns:
        An integer value within a given range.

    """
    return draw(st.integers(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_LONG_TYPE)
@st.composite
def long_strategy(draw: st.DrawFn) -> int:
    """Generate a long value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A long value.

    """
    return draw(st.integers(min_value=-(2**63), max_value=2**63 - 1))


@register_strategy(PYSPARK_SHORT_TYPE)
@st.composite
def short_strategy(draw: st.DrawFn) -> int:
    """Generate a short value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A short value.

    """
    return draw(st.integers(min_value=-(2**15), max_value=2**15 - 1))


@register_strategy(PYSPARK_STRING_TYPE)
@st.composite
def string_strategy(
    draw: st.DrawFn, min_size: int = 0, max_size: Optional[int] = None
) -> str:
    """Generate a string value with a given length.

    Args:
        draw: The Hypothesis draw function.
        min_size: The minimum length of the string.
        max_size: The maximum length of the string.

    Returns:
        A string value with a given length

    """
    return draw(st.text(min_size=min_size, max_size=max_size))


@st.composite
def update_pandas_df_strategy(
    draw: st.DrawFn,
    pandas_df: pd.DataFrame,
    columns: list[dict],
):
    """Apply custom strategies to a Pandas DataFrame based on the custom data.

    Args:
        draw: The Hypothesis draw function.
        pandas_df: The Pandas DataFrame to apply custom strategies to.
        columns: The custom data columns to apply strategies to.

    Returns:
        A Pandas DataFrame with custom strategies applied.

    """
    pandas_df_copy = pandas_df.copy()
    number_of_rows = len(pandas_df_copy)

    for column in columns:
        column_dtype = column.get(CUSTOM_DATA_TYPE_KEY)
        strategy = snowpark_strategies.get(column_dtype, None)

        if strategy is None:
            raise ValueError(
                f"Unsupported custom strategy for data type: {column_dtype}"
            )

        strategy_kwargs = {}

        if column_dtype == PYSPARK_ARRAY_TYPE:
            strategy_kwargs = {
                "dtype": column.get(CUSTOM_DATA_VALUE_TYPE_KEY),
                "min_size": column.get(CUSTOM_DATA_MIN_SIZE_KEY, 0),
                "max_size": column.get(CUSTOM_DATA_MAX_SIZE_KEY, None),
            }
        elif column_dtype == PYSPARK_BINARY_TYPE:
            strategy_kwargs = {
                "min_size": column.get(CUSTOM_DATA_MIN_SIZE_KEY, 0),
                "max_size": column.get(CUSTOM_DATA_MAX_SIZE_KEY, None),
            }

        column_name = column.get(CUSTOM_DATA_NAME_KEY)
        pandas_df_copy[column_name] = draw(
            st.lists(
                strategy(**strategy_kwargs),
                min_size=number_of_rows,
                max_size=number_of_rows,
            )
        )

    return pandas_df_copy
